#include <QtCore/QtGlobal>
#include <QtCore/QDebug>
#include <QtCore/QMutex>
#include <QtCore/QWaitCondition>
#include <iostream>
#include "circularbuffertracker.h"

CircularBufferTracker::CircularBufferTracker()
    :capacity(0), history(0), states(NULL), finish(false)
{
    qFatal("CircularBufferTracker: default constructor not suppported.");
}


CircularBufferTracker::CircularBufferTracker(unsigned int capacity,
                                             unsigned int history)
    : capacity(capacity),
      history(history),
      states(new RecordState[capacity]),
      finish(false)
{
    if (!(capacity > history)) {
        qFatal("CircularBufferTracker: object must have capacity (%d) "
               "greater than history (%d).", capacity, history);
    }
    for (std::ptrdiff_t i = 0; i < static_cast<std::ptrdiff_t>(this->capacity); ++i) {
        states[i] = EMPTY;
    }
    this->read       = 0;
    this->written    = 0;
    this->past       = 0;
    this->readIndex  = 0;
    this->writeIndex = 0;
    this->mutex = new QMutex ();
    this->canRead = new QWaitCondition ();
    this->canWrite = new QWaitCondition ();
}

CircularBufferTracker::~CircularBufferTracker()
{
    delete [] this->states;
    delete this->mutex;
    delete this->canRead;
    delete this->canWrite;
}

std::ptrdiff_t CircularBufferTracker::startWrite()
{
    this->mutex->lock ();
    while ( !( (this->written < this->capacity)
               || (this->past > this->history) ) ) {
        this->canWrite->wait (this->mutex);
    }
    std::ptrdiff_t const wi = this->writeIndex;
    //qDebug () << "startWrite" << this->toString ().c_str ();
    this->mutex->unlock ();
    return wi;
}

void CircularBufferTracker::endWrite(std::ptrdiff_t writtenIndex)
{
    if (this->writeIndex != writtenIndex) {
        qFatal ("CircularBufferTracker: Wrong write index!");
    }
    this->mutex->lock ();
    this->written++;

    // If we are already overwriting slots ...
    if (this->written > this->capacity) {
        this->past--; // ... we are erasing history!
    }

    this->states[this->writeIndex] = UNPROCESSED;

    this->writeIndex++;
    if (this->writeIndex >= static_cast<std::ptrdiff_t>(this->capacity)) {
        this->writeIndex -= this->capacity;
    }
    //qDebug () << "endWrite" << this->toString ().c_str ();
    this->mutex->unlock ();
    this->canRead->wakeAll ();
}

std::ptrdiff_t CircularBufferTracker::startRead(unsigned int *pastElements)
{
    this->mutex->lock ();
    while ( !(written > read) ) {
        // We enter this part of the loop only when the buffer is empty
        if (this->finish) {
            // We intentionally leave the mutex locked since this buffer is
            // no longer valid.
            this->mutex->unlock ();
            return -1; // We signal that there is no more valid data
        }
        this->canRead->wait (this->mutex);
    }
    this->states[this->readIndex] = CURRENT;
    std::ptrdiff_t const ri = this->readIndex;
    if (NULL != pastElements) {
        *pastElements = this->past;
    }
    //qDebug () << "startRead" << this->toString ().c_str ();
    //qDebug () << this->past << " past elements available";
    this->mutex->unlock ();
    return ri;
}

void CircularBufferTracker::endRead(std::ptrdiff_t readIndex)
{
    //qDebug () << "endRead";
    if (this->readIndex != readIndex) {
        qFatal ("CircularBufferTracker: Wrong read index!");
    }
    this->mutex->lock ();
    this->read++; // Increase read counter
    this->past++; // The just read element is now history

    this->states[this->readIndex] = PAST;

    this->readIndex++;
    if (this->readIndex >= static_cast<std::ptrdiff_t>(this->capacity)) {
        this->readIndex -= this->capacity;
    }
    //qDebug () << "endRead" << this->toString ().c_str ();
    this->mutex->unlock ();
    this->canWrite->wakeAll ();
}

std::string CircularBufferTracker::toString() const
{
    //qDebug () << "toString";
    char *charStates = new char[this->capacity + 1];
    charStates[this->capacity] = 0; // Terminate string
    //this->mutex.lock ();
    for (std::ptrdiff_t i = 0; i < static_cast<std::ptrdiff_t>(this->capacity); ++i) {
        charStates[i] = static_cast<char>(this->states[i]);
    }
    //this->mutex.unlock ();
    std::string stringStates(charStates);
    delete [] charStates;
    return stringStates;
}

unsigned int CircularBufferTracker::getAvailablePast() const
{
    return this->past;
}


void testCircularBufferTracker (CircularBufferTracker &cbt, char const operations[]) {
    bool readStarted = false;
    bool writeStarted = false;
    bool terminate = false;
    std::ptrdiff_t readIndex = -1;
    std::ptrdiff_t writeIndex = -1;
    for (std::ptrdiff_t i = 0; false == terminate; ++i) {
        switch (operations[i]) {
        case 'r':
            if (readStarted) {
                qFatal("Illegal read start attempt!");
            } else {
                readStarted = true;
            }
            readIndex = cbt.startRead ();
            break;
        case 'R':
            readStarted = false;
            cbt.endRead (readIndex);
            readIndex = -1;
            break;
        case 'w':
            if (writeStarted) {
                qFatal("Illegal write start attempt!");
            } else {
                writeStarted = true;
            }
            writeIndex = cbt.startWrite ();
            break;
        case 'W':
            writeStarted = false;
            cbt.endWrite (writeIndex);
            writeIndex = -1;
            break;
        case '.':
            terminate = true;
            break;
        default:
            qFatal("Unsupported operation: '%c'", operations[i]);
        }
        std::cerr << "Operation: " << operations[i] << ". States: " <<  cbt.toString () << "\n";
    }
}

unsigned int CircularBufferTracker::getCapacity() const
{
    return capacity;
}

void CircularBufferTracker::notifyEndOfStream()
{
    this->finish = true;
}
