#ifndef CIRCULARBUFFERTRACKER_H
#define CIRCULARBUFFERTRACKER_H

#include <cstddef>
#include <string>


class QMutex;
class QWaitCondition;

// Implements a tracker for managing a circular buffer that optionally maintains
// a number of records after being read for reference purposes.
// It is intended to be used by only ONE producer and ONE consumer, since
// the read and write operations are not atomic.
// Since we will be using it for arbitrating the access to CUDA layered
// surfaces/textures (i.e. the data is not actually contained here) we need
// to adopt a convention: increasing the pointer goes into the future,
// decreasing it goes into the past.
class CircularBufferTracker
{
public:
    enum RecordState {EMPTY       = 'E',
                      CURRENT     = 'C',
                      PAST        = 'P',
                      UNPROCESSED = 'U'};

    CircularBufferTracker(unsigned int capacity, unsigned int history = 0);
    virtual ~CircularBufferTracker ();

    // Returns the index in which the next insertion must take place. Blocks if no free slot is available
    std::ptrdiff_t startWrite ();

    // Notifies the tracker that the insertion is completed. Parameter provides
    // a sanity check, maybe it is unnecessary
    void endWrite (std::ptrdiff_t writtenIndex);

    // Returns the index of the current element to be processed. Blocks if no free slot is available
    // and returns -1 if the producer has signaled the end of the data stream.
    // If pastElements is not NULL, it will contain the number of available
    // past elements upon return.
    std::ptrdiff_t startRead (unsigned int *pastElements = NULL);

    // Notifies the tracker that the read and associated processing is completed.
    // see endWrite.
    void endRead (std::ptrdiff_t readIndex);

    // Returns a string representation of the states of each slot in the buffer.
    // Note that it is not thread-safe by itself. If the object is not frozen
    // the output can be incoherent.
    std::string toString () const;

    unsigned int getAvailablePast () const;

    unsigned int getCapacity () const;

    void notifyEndOfStream ();

private:
    CircularBufferTracker(); // We don't want to allow this

    unsigned int const capacity; // Size of the buffer
    unsigned int const history; // Minimum number of past records to keep
    RecordState *const states; // Map of the buffer
    unsigned int       read; // Total number of read (processed) elements
    unsigned int       written; // Total number of written elements
    unsigned int       past; // Total number of historic records past currently read
    std::ptrdiff_t     readIndex; // Index of the current record
    std::ptrdiff_t     writeIndex; // Index of the next record to be written
    QMutex            *mutex;
    QWaitCondition    *canWrite;
    QWaitCondition    *canRead;
    bool               finish;
};


// To test the buffer. The possible operations are:
// r -> Start read
// R -> End read
// w -> Start write
// W -> End write
// . -> End of string
// It is illegal to start a new operation of the same kind as another that has
// not ended yet, and "operations" must end with "."
void testCircularBufferTracker (CircularBufferTracker &cbt, char const operations[]);

#endif // CIRCULARBUFFERTRACKER_H
