#include <QtCore/QDebug>
#include <cuda_runtime.h>
#include <FreeImagePlus.h>
#include <cassert>
#include "inputstage.h"
#include "auxcuda.h"
#include "inputstagecuda.h"


template <typename ELEMENT_TYPE>
void InputStage<ELEMENT_TYPE>::initialize()
{
    this->inputImages = getImageList(this->ioConfig.inputDir);
    fipImage inImage;
    fipImage channel;
    assert(inImage.load(this->inputImages.front().native().c_str()));

    if (352 != inImage.getWidth()) {
        this->imageDim.width = 352;
        this->imageDim.height = static_cast<unsigned int>(0.5f + (inImage.getHeight() * (352.0f / inImage.getWidth()))); // Rounded scaled size
        std::cout << this->imageDim.width << "x" << this->imageDim.height << std::endl;
    } else {
        this->imageDim.width = inImage.getWidth ();
        this->imageDim.height = inImage.getHeight ();
    }
    this->imageDim.channels = 3; // TODO :  This should be detected from the image if ever we use multiband images.
    assert (inImage.getChannel (channel, FICC_RED));
    assert (channel.rescale(this->imageDim.width, this->imageDim.height, FILTER_BOX)); // We just want to know the scanline dimensions

    this->originalPinnedChannels = new unsigned char * [this->imageDim.channels];

    for (unsigned int ch = 0; ch < this->imageDim.channels; ++ch) {
        // We use pinned memory because it allows faster transfer to the device
        CUDA_SAFE_CALL (cudaMallocHost (&this->originalPinnedChannels[ch], channel.getScanWidth () * channel.getHeight ()));
    }

    cudaChannelFormatDesc const ucharFormat = makeCudaChannelFormatDesc<unsigned char> (); // TODO : The actual type should be detected if ever we use higher bit depth input images
    CUDA_SAFE_CALL (cudaMalloc3DArray (&this->originalInputArray,
                                       &ucharFormat,
                                       make_cudaExtent(this->imageDim.width,
                                                       this->imageDim.height,
                                                       this->imageDim.channels),
                                       cudaArrayLayered | cudaArraySurfaceLoadStore));
    const cudaChannelFormatDesc floatFormat = makeCudaChannelFormatDesc<float>();
    CUDA_SAFE_CALL (cudaMalloc3DArray (&this->normalizedFloatArray,
                                       &floatFormat,
                                       make_cudaExtent(this->imageDim.width,
                                                       this->imageDim.height,
                                                       this->imageDim.channels),
                                       cudaArrayLayered | cudaArraySurfaceLoadStore));
    this->nextStageArray = this->next->initializeBuffer (this->imageDim);
    assert (NULL != this->nextStageArray);
    bindInputSurfaces (this->originalInputArray, this->normalizedFloatArray, this->nextStageArray);
}

template <typename ELEMENT_TYPE>
void InputStage<ELEMENT_TYPE>::run()
{
//    qDebug ("Input stage starting\n");



    this->initialize();


    cudaStream_t stream;
    CUDA_SAFE_CALL (cudaStreamCreate(&stream));


    fipImage inImage;
    fipImage *inChannels = new fipImage[this->imageDim.channels];
    QString relativeFilename;
    cudaMemcpy3DParms memcpyParams;
    memset(&memcpyParams, 0, sizeof (memcpyParams));
    memcpyParams.extent = make_cudaExtent (imageDim.width, imageDim.height, 1); // Docs say width should be bytes, but it's actually elements
    memcpyParams.kind   = cudaMemcpyHostToDevice;

    for (PathVector::const_iterator i = this->inputImages.begin(); i != this->inputImages.end(); ++i) {
        // Load image and separate it into channels
        qDebug () << "Opening:" << i->native().c_str();

        assert (inImage.load (i->native().c_str()));
        assert (inImage.rescale(this->imageDim.width, this->imageDim.height, FILTER_BILINEAR));
        assert (inImage.splitChannels (inChannels[0], inChannels[1], inChannels[2])); // TODO : Again, 3-channel specific

        // Copy each channel of the image into an array for filtering and conversion into the desired format
        for (unsigned int ch = 0; ch < this->imageDim.channels; ++ch) {
            // First copy into pinned memory
            memcpy(this->originalPinnedChannels[ch], inChannels[ch].accessPixels(), inChannels[ch].getScanWidth() * inChannels[ch].getHeight());
            // Then to the array
            memcpyParams.dstPos = make_cudaPos (0, 0, ch);
            memcpyParams.srcPtr = make_cudaPitchedPtr (this->originalPinnedChannels[ch],
                                                       inChannels[ch].getScanWidth (),
                                                       inChannels[ch].getWidth (),
                                                       inChannels[ch].getHeight ());
            memcpyParams.dstArray = this->originalInputArray;
            CUDA_SAFE_CALL (cudaMemcpy3DAsync(&memcpyParams, stream));
        }
        //CUDA_SAFE_CALL (cudaStreamSynchronize (stream));


        std::ptrdiff_t const writePtr = this->next->startWrite ();

        switch (this->procConfig.colourSpace) {
        case ProcessingConfig::RGB:
            normalizeAndCopyRGB<ELEMENT_TYPE> (this->imageDim.width,
                                               this->imageDim.height,
                                               writePtr,
                                               stream);
            break;
        case ProcessingConfig::RnGnGradS:
            normalizeAndCopyRnGnGradS<ELEMENT_TYPE> (this->imageDim.width,
                                                     this->imageDim.height,
                                                     writePtr,
                                                     stream);
            break;
        default:
            qFatal("Unknown colour space!");
        }
        this->next->endWrite (writePtr);
    }
    this->next->notifyEndOfStream ();
}

template <typename ELEMENT_TYPE>
InputStage<ELEMENT_TYPE>::InputStage(const ProcessingConfig &procConfig, const IOConfig & config, Stage *next)
    : ioConfig(config), procConfig(procConfig)
{
    this->next = next;
}

template <typename ELEMENT_TYPE>
InputStage<ELEMENT_TYPE>::~InputStage()
{
    unbindInputSurfaces ();
    CUDA_SAFE_CALL (cudaFreeArray (this->originalInputArray));
    this->originalInputArray = NULL;
    CUDA_SAFE_CALL (cudaFreeArray (this->normalizedFloatArray));
    this->normalizedFloatArray = NULL;
    for (unsigned int ch = 0; ch < this->imageDim.channels; ++ch) {
        // We use pinned memory because it allows faster transfer to the device
        CUDA_SAFE_CALL (cudaFreeHost (this->originalPinnedChannels[ch]));
    }
    delete [] this->originalPinnedChannels;
    qDebug ("Salimos del destructor de Input Stage");
}

// Explicit instantiation
template class InputStage<unsigned char>;
template class InputStage<unsigned short>;
