#include <surface_types.h>
#include <surface_functions.h>
#include "inputstagecuda.h"
#include "auxcuda.h"
#include "scalarutils.h"

surface<void, cudaSurfaceType2DLayered> normalizedSurface;
surface<void, cudaSurfaceType2DLayered> nextStageSurface;

texture<unsigned char, cudaTextureType2DLayered, cudaReadModeElementType>  ucharTexture;
texture<float, cudaTextureType2DLayered, cudaReadModeElementType>  floatTexture;


void bindInputSurfaces (cudaArray *originalInputArray,
                        cudaArray *normalizedFloatArray,
                        cudaArray *nextStageArray)
{
    ucharTexture.addressMode[0] = cudaAddressModeClamp;
    ucharTexture.addressMode[1] = cudaAddressModeClamp;
    ucharTexture.filterMode     = cudaFilterModePoint;
    ucharTexture.normalized     = 0;
    CUDA_SAFE_CALL (cudaBindTextureToArray (ucharTexture, originalInputArray));

    CUDA_SAFE_CALL (cudaBindSurfaceToArray (normalizedSurface, normalizedFloatArray));

    floatTexture.addressMode[0] = cudaAddressModeBorder;
    floatTexture.addressMode[1] = cudaAddressModeBorder;
    floatTexture.filterMode     = cudaFilterModePoint;
    floatTexture.normalized     = 0;
    CUDA_SAFE_CALL (cudaBindTextureToArray (floatTexture, normalizedFloatArray));

    CUDA_SAFE_CALL (cudaBindSurfaceToArray (nextStageSurface, nextStageArray));
}

void unbindInputSurfaces () {
    CUDA_SAFE_CALL (cudaUnbindTexture (ucharTexture));
    CUDA_SAFE_CALL (cudaUnbindTexture (floatTexture));
}


template <typename T>
__global__ void kernelUnalteredRGB (int width, int height, int offset) {
    const int x = blockIdx.x * blockDim.x + threadIdx.x;
    const int y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= width) || (y >= height)) { return; }

    unsigned char const r = tex2DLayered (ucharTexture, x, y, 0);
    unsigned char const g = tex2DLayered (ucharTexture, x, y, 1);
    unsigned char const b = tex2DLayered (ucharTexture, x, y, 2);
    const float rF = IntFloatConverter<unsigned char>::toFloat (r);
    const float gF = IntFloatConverter<unsigned char>::toFloat (g);
    const float bF = IntFloatConverter<unsigned char>::toFloat (b);
    surf2DLayeredwrite<T> (IntFloatConverter<T>::fromFloat (rF), nextStageSurface, sizeof (T) * x, y, offset + 0);
    surf2DLayeredwrite<T> (IntFloatConverter<T>::fromFloat (gF), nextStageSurface, sizeof (T) * x, y, offset + 1);
    surf2DLayeredwrite<T> (IntFloatConverter<T>::fromFloat (bF), nextStageSurface, sizeof (T) * x, y, offset + 2);
}




template <typename T>
__global__ void kernelLowpassFilterGradI (int width, int height, int offset) {
    const int x = blockIdx.x * blockDim.x + threadIdx.x;
    const int y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= width) || (y >= height)) { return; }

    const float gaussianMask[3][3] = {{0.011343736558495f, 0.083819505802211f, 0.011343736558495f},
                                      {0.083819505802211f, 0.619347030557180f, 0.083819505802211f},
                                      {0.011343736558495f, 0.083819505802211f, 0.011343736558495f}};

    float filteredGradS = 0.0f;
    
    for (int i = -1; i <= 1; ++i) {
        for (int j = -1; j <= 1; ++j) {
            float const gradSSample = tex2DLayered (floatTexture, x + j, y + i, 2); // gradS layer
            filteredGradS += gradSSample * gaussianMask[i+1][j+1];
        }
    }
    surf2DLayeredwrite<T> (IntFloatConverter<T>::fromFloat (tex2DLayered (floatTexture, x, y, 0)), nextStageSurface, sizeof (T) * x, y, offset + 0);
    surf2DLayeredwrite<T> (IntFloatConverter<T>::fromFloat (tex2DLayered (floatTexture, x, y, 1)), nextStageSurface, sizeof (T) * x, y, offset + 1);
    surf2DLayeredwrite<T> (IntFloatConverter<T>::fromFloat (filteredGradS), nextStageSurface, sizeof (T) * x, y, offset + 2);
    // TODO : Esto es para debuggear y poder comparar con lo del año pasado. Al acabar se puede retirar sin problemas
//    surf2DLayeredwrite<T> (IntFloatConverter<T>::fromFloat (tex2DLayered (floatTexture, x, y, 0)), nextStageSurface, sizeof (T) * x, y, offset + 2);
//    surf2DLayeredwrite<T> (IntFloatConverter<T>::fromFloat (tex2DLayered (floatTexture, x, y, 1)), nextStageSurface, sizeof (T) * x, y, offset + 1);
//    surf2DLayeredwrite<T> (IntFloatConverter<T>::fromFloat (filteredGradS), nextStageSurface, sizeof (T) * x, y, offset + 0);
}


__global__ void kernelRGBToNormalizedRGGradI (int width, int height)
{
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;

    if ((0 > x) || (0 > y) || (x >= width) || (y >= height)) { return; }

    int gx = 0, gy = 0;
    int const sobelVerticalMask[3][3] = { 1,  2,  1,
                                          0,  0,  0,
                                         -1, -2, -1};

    float floatResults[3];

    for (int i = -1; i <= 1; ++i) {
        for (int j = -1; j <= 1; ++j) {
            unsigned char const r = tex2DLayered (ucharTexture, x + j, y + i, 0);
            unsigned char const g = tex2DLayered (ucharTexture, x + j, y + i, 1);
            unsigned char const b = tex2DLayered (ucharTexture, x + j, y + i, 2);

            int const S = static_cast<int>(r)
                          + static_cast<int>(g)
                          + static_cast<int>(b);
            gy += S * sobelVerticalMask[i+1][j+1];
            gx += S * sobelVerticalMask[j+1][i+1];

            // The central pixel
            if ((0 == i) && (0 == j)) {
                floatResults[0] = static_cast<float>(r) / static_cast<float>(S);
                floatResults[1] = static_cast<float>(g) / static_cast<float>(S);
            }
        }
    }

    floatResults[2] = sqrtf(static_cast<float>(gx * gx + gy * gy)) / static_cast<float>(8   // sum(abs(sobel Weights))
                                                                                        * 3 // Number of channels that make S
                                                                                        * UCHAR_MAX);
#pragma unroll
    for (int ch = 0 ; ch < 3 ; ++ch) {
        surf2DLayeredwrite<float> (floatResults[ch],
                        normalizedSurface,
                        sizeof (float) * x, y, ch);
    }
}


template<typename ELEMENT_TYPE>
void normalizeAndCopyRnGnGradS (int width, int height, int offset, cudaStream_t stream)
{
    const int blockEdge = 16;
    const dim3 blockSize(blockEdge, blockEdge);
    dim3 const gridSize (1 + ( width - 1) / blockSize.x,
                         1 + (height - 1) / blockSize.y);
    kernelRGBToNormalizedRGGradI<<<gridSize, blockSize, 0, stream>>>(width, height);
    //CUDA_SAFE_CALL (cudaGetLastError());
    kernelLowpassFilterGradI<ELEMENT_TYPE><<<gridSize, blockSize, 0, stream>>>(width, height, offset);
    CUDA_SAFE_CALL (cudaStreamSynchronize (stream));
    CUDA_SAFE_CALL (cudaGetLastError());
}

template void normalizeAndCopyRnGnGradS<unsigned char>(int width, int height, int offset, cudaStream_t stream);
template void normalizeAndCopyRnGnGradS<unsigned short>(int width, int height, int offset, cudaStream_t stream);


template <typename ELEMENT_TYPE>
void normalizeAndCopyRGB (int width, int height, int offset, cudaStream_t stream) {
    const int blockEdge = 16;
    const dim3 blockSize(blockEdge, blockEdge);
    dim3 const gridSize (1 + ( width - 1) / blockSize.x,
                         1 + (height - 1) / blockSize.y);
    kernelUnalteredRGB<ELEMENT_TYPE><<<gridSize, blockSize, 0, stream>>>(width, height, offset);
    CUDA_SAFE_CALL (cudaStreamSynchronize (stream));
    CUDA_SAFE_CALL (cudaGetLastError());
}

template void normalizeAndCopyRGB<unsigned char> (int width, int height, int offset, cudaStream_t stream);
template void normalizeAndCopyRGB<unsigned short> (int width, int height, int offset, cudaStream_t stream);


