#include <cuda_runtime.h>
#include <FreeImagePlus.h>
#include <cassert>
#include <QtCore/QDebug>
#include <QtCore/QCoreApplication>
#include <boost/filesystem.hpp>
#include "outputstage.h"
#include "auxcuda.h"
#include "filesystem.h"

OutputStage::OutputStage(const IOConfig & config, Stage * /* next */)
    : tracker(config.outputBufferSize, 0), // No past frames are necessary
      config(config)
{

}

OutputStage::~OutputStage()
{

}

std::ptrdiff_t OutputStage::startWrite()
{
    return this->tracker.startWrite () * this->imageDim.channels;
}

void OutputStage::endWrite(std::ptrdiff_t written)
{
    return this->tracker.endWrite (written / this->imageDim.channels);
}

void OutputStage::notifyEndOfStream()
{
    this->tracker.notifyEndOfStream ();
}

cudaArray *OutputStage::initializeBuffer(ImageDimensions const &dim)
{
    this->imageDim = dim;
    return this->initializeBuffer ();
}



cudaArray *OutputStage::initializeBuffer(unsigned int width, unsigned int height, unsigned int channels)
{
    this->imageDim.width = width;
    this->imageDim.height = height;
    this->imageDim.channels = channels;
    return this->initializeBuffer ();
}

OutputStage::OutputStage()
    : tracker(1) // Needed to compile because CBT does not accept default constructor
{
}

cudaArray *OutputStage::initializeBuffer()
{
    cudaChannelFormatDesc const floatFormat = makeCudaChannelFormatDesc<float>();
    CUDA_SAFE_CALL (cudaMalloc3DArray (&this->outputImagesArray,
                                       &floatFormat,
                                       make_cudaExtent(this->imageDim.width,
                                                       this->imageDim.height,
                                                       this->imageDim.channels * this->tracker.getCapacity ()),
                                       cudaArrayLayered | cudaArraySurfaceLoadStore));
    return this->outputImagesArray;
}

void OutputStage::run()
{
//    qDebug ("Output stage starting\n");
    boost::filesystem::path outputDir(this->config.outputDir);
    if (boost::filesystem::exists(outputDir)) {
        assert(boost::filesystem::is_directory(outputDir));
        boost::filesystem::remove_all(outputDir);
    }
    assert(boost::filesystem::create_directories(outputDir));

    const PathVector inputImages = getImageList(this->config.inputDir);

    cudaStream_t stream;
    CUDA_SAFE_CALL (cudaStreamCreate(&stream));

    fipImage image;
    assert(image.load(inputImages.front().native().c_str()));
    const unsigned int originalWidth = image.getWidth();
    const unsigned int originalHeight = image.getHeight();
    //fipImage pngImage;

    QString relativeFilename;
    cudaMemcpy3DParms memcpyParams;
    memset(&memcpyParams, 0, sizeof (memcpyParams));
    memcpyParams.kind   = cudaMemcpyDeviceToHost;

    for (PathVector::const_iterator i = inputImages.begin(); i != inputImages.end(); ++i) {
//        // Load image and separate it into channels
        std::ptrdiff_t const readPtr = this->tracker.startRead ();
        assert (image.setSize (FIT_FLOAT, this->imageDim.width, this->imageDim.height, 32));
        //assert (pngImage.setSize (FIT_BITMAP, this->imageDim.width, this->imageDim.height, 8));
        memcpyParams.extent = make_cudaExtent (this->imageDim.width, this->imageDim.height, 1); // Docs say width should be bytes, but it's actually elements
        memcpyParams.srcPos = make_cudaPos (0, 0, readPtr);
        memcpyParams.dstPtr = make_cudaPitchedPtr (image.accessPixels (),
                                                   image.getScanWidth (),
                                                   image.getWidth (),
                                                   image.getHeight ());
        memcpyParams.srcArray = this->outputImagesArray;
        CUDA_SAFE_CALL (cudaMemcpy3DAsync(&memcpyParams, stream));
        CUDA_SAFE_CALL (cudaStreamSynchronize (stream));

//        qDebug ("Consumer ending read %06d", i);
        this->tracker.endRead (readPtr);
        if (this->config.savePNG) {
            for (unsigned int y = 0 ; y < image.getHeight(); ++y) {
                float * line = reinterpret_cast<float *>(image.getScanLine(y));
                for (unsigned int x = 0; x < image.getWidth(); ++x) {
                    line[x] *= 255.0f;
                }
            }
            assert(image.convertToType(FIT_BITMAP, false));
        }
        assert(image.rescale(originalWidth, originalHeight, FILTER_BILINEAR));
        if ("Aa" != i->filename().native().substr(0, 2)) {
            assert(image.save((outputDir / i->filename()).native().c_str()));
        }
    }
}

