#include "particlefiltercuda.h"

/// NOTA: Podría parecer buena idea que en el caso de computeParticleCoverageMask podríamos salir del bucle
/// tan pronto como una partícula nos cubra el píxel, no hace falta recorrer la lista completa. Además, es de
/// esperar que esto tenga poca divergencia porque los píxeles cubiertos por la misma partícula tienen una
/// relación de vecindario y, por tanto, tienen muchas probabilidades de pertenecer al mismo warp.
/// Sin embargo, empíricamente se demuestra que no hay diferencia alguna, imagino que en parte debido al
/// hecho de que las partículas del array están ordenadas por regiones, no barajadas. Así, a la región que le
/// toquen las últimas partículas le va a tocar recorrer todo el array igualmente. De todos modos debe de ser
/// una mierdecilla en el cómputo global.
///
/// Ni eso. Lo que pasa es que en realidad estamos haciendo esto con vectores de estado, no partículas, que son pocos y cobardes.
__device__ unsigned int getNumberOfParticlesCoveringPixel (int x, int y,
                                                           unsigned int numParticles,
                                                           const Particle * __restrict__ particles) {
    unsigned int covered = 0;

    for (unsigned int p = 0; p < numParticles; ++p) {
        const float xNormDist = (static_cast<float>(x) - particles[p].centreX) / particles[p].radiusX;
        const float yNormDist = (static_cast<float>(y) - particles[p].centreY) / particles[p].radiusY;
        covered += ((xNormDist * xNormDist + yNormDist * yNormDist) <= 1) ? 1 : 0;
    }
    return covered;
}

__global__ void computeParticleCoverageMask (ImageDimensions                imageDim,
                                             unsigned int                   numParticles,
                                             const Particle  * __restrict__ particles,
                                             bool            * __restrict__ coverageMask) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }
    const unsigned int numberOfParticlesCoveringCurrentPixel = getNumberOfParticlesCoveringPixel (x, y, numParticles, particles);
    *displacePtr3D(coverageMask, imageDim.width, imageDim.height, x, y, 0) = (0 < numberOfParticlesCoveringCurrentPixel);
}

__global__ void computeParticleCoverageParity (ImageDimensions                imageDim,
                                               unsigned int                   numParticles,
                                               const Particle  * __restrict__ particles,
                                               unsigned char   * __restrict__ results) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    const unsigned int  numberOfParticlesCoveringCurrentPixel = getNumberOfParticlesCoveringPixel (x, y, numParticles, particles);
    const unsigned char coverageParity = (0 < numberOfParticlesCoveringCurrentPixel)
                                         ? (1 + numberOfParticlesCoveringCurrentPixel % 2)
                                         : 0;
    *displacePtr3D(results, imageDim.width, imageDim.height, x, y, 0) = coverageParity;
}


__global__ void computeParticleFGPriorProbability (ImageDimensions                imageDim,
                                            unsigned int                   numParticles,
                                            const Particle  * __restrict__ particles,
                                            float           * __restrict__ probability) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }
    float pxProb = 0.0f;
    unsigned int particlesCoveringCurrentPx = 0;
    const float minusLn2 = -0.69314718055994529f;
    for (unsigned int p = 0; p < numParticles; ++p) {
        const float xNormDist = (static_cast<float>(x) - particles[p].centreX) / particles[p].radiusX;
        const float yNormDist = (static_cast<float>(y) - particles[p].centreY) / particles[p].radiusY;
        const float squaredNormDist = xNormDist * xNormDist + yNormDist * yNormDist;
        if (1 >= squaredNormDist) { // we don't want to compute a lot of tiny exponentials
//            pxProb += __expf(minusLn2 * squaredNormDist);
            pxProb += __expf(-5.9168 * squaredNormDist);
            ++particlesCoveringCurrentPx;
        }
    }
    Pointer2D<float> probPtr(probability, imageDim.width);
//    probPtr(x,y) = (particlesCoveringCurrentPx > 3) ? pxProb / static_cast<float>(particlesCoveringCurrentPx) : 0.5f;
    probPtr(x,y) = (particlesCoveringCurrentPx > 4) ? 0.5f * (1.0f + (pxProb / static_cast<float>(particlesCoveringCurrentPx))) : 0.5f;
}

__global__ void computeConstantFGPriorProbability (ImageDimensions                imageDim,
                                            float           * __restrict__ probability) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    Pointer2D<float> probPtr(probability, imageDim.width);
    probPtr(x,y) = 0.5f;
}





/// Writes a 1 in coveredRegions[i] if region i-th is covered according to coverageMask
/// coveredRegions needs to be uint because SM 2.0 doesn't support atomic boolean operations
/// TODO: FIXME: Since final assignation is non-atomic, maybe we can use bool after all
extern __shared__ unsigned int privateCoveredRegions[];
__global__ void determineCoveredRegions (      unsigned int                   numMaxRegions,
                                               ImageDimensions                imageDim,
                                         const PixelInRegion   * __restrict__ labelledPixels,
                                               unsigned int                   numberOfLabelledPixels,
                                         const bool            * __restrict__ coverageMask,
                                               unsigned int    * __restrict__ coveredRegions) {
    const unsigned int globalThreadId = blockIdx.x * blockDim.x + threadIdx.x;

    //__shared__ unsigned int privateCoveredRegions[numMaxRegions + 1]; // Label 0 doesn't count
//    for (unsigned int i = threadIdx.x; i <= numMaxRegions; i += blockDim.x) {
//        privateCoveredRegions[i] = 0;
//    }
//    __syncthreads();

    if (globalThreadId < numberOfLabelledPixels) {
        PixelInRegion pix = labelledPixels[globalThreadId];
        if (*displacePtr3D(coverageMask, imageDim.width, imageDim.height, pix.col, pix.row, 0)) {
//            const unsigned int oldCoveredValue = atomicExch(&privateCoveredRegions[pix.label], 1);
//            if (0 == oldCoveredValue) {
                coveredRegions[pix.label] = 1;
//            }
        }
    }
}

__global__ void weighParticles(ImageDimensions                    imageDim,
                               float                              likelihoodBeta,
                               const unsigned char * __restrict__ bgProbability,
                               const PixelInRegion * __restrict__ labelledPixels,
                               unsigned int                       numberOfLabelledPixels,
                               Particle            * __restrict__ particles,
                               unsigned int                       numParticles) {
    /// Cada hilo será responsable de una partícula e iterará por todos los píxeles. De esta forma no es necesario hacer reducciones, cosa que sí
    /// sería necesaria en caso de iterar al revés. Además esto posibilitará usar números enormes de partículas con poca penalización
    const unsigned int partIdx = blockIdx.x * blockDim.x + threadIdx.x;

    if (partIdx >= numParticles) {
        return;
    }
    Particle & thisParticle = particles[partIdx];
    const float partRowRadiusSquared = thisParticle.radiusY * thisParticle.radiusY;
    const float partColRadiusSquared = thisParticle.radiusX * thisParticle.radiusX;
    const float partRowSigmaSquared = partRowRadiusSquared * 0.084505138; // 1/3.44^2 = 0.084505138
    const float partColSigmaSquared = partColRadiusSquared * 0.084505138; // 1/3.44^2 = 0.084505138

    unsigned int j_in = 0;
    unsigned int j    = 0;
    float        f3   = 1.0f;
//    float alternatef3Sum = 0.0f;
//    float alternatef3Weights = 0.000001f;
    const float  rho  = 1.0f;
    for (unsigned int pxIdx = 0; pxIdx < numberOfLabelledPixels; ++pxIdx) {
        const PixelInRegion & currentPixel = labelledPixels[pxIdx];
        if (currentPixel.label == thisParticle.associatedLabel) {
            ++j;
            const float rowDelta = static_cast<float>(currentPixel.row) - thisParticle.centreY;
            const float colDelta = static_cast<float>(currentPixel.col) - thisParticle.centreX;
            const float rowDeltaSquared = rowDelta * rowDelta;
            const float colDeltaSquared = colDelta * colDelta;
            if ((1 >= (rowDeltaSquared / partRowRadiusSquared) + (colDeltaSquared / partColRadiusSquared))) {
                ++j_in;
            }
            const float g = __expf(-0.5 * ((rowDeltaSquared / partRowSigmaSquared) + (colDeltaSquared / partColSigmaSquared)));
//            if (1.0 < g) {
//                printf("JAAAAAAAAAAAAAAARL!\n", partIdx, thisParticle.associatedLabel);
//            }
            const unsigned char prBG = *(displacePtr3D<unsigned char>(bgProbability,
                                                                      imageDim.width,
                                                                      imageDim.height,
                                                                      currentPixel.col, currentPixel.row, 0));
            const float prFG = static_cast<float>(255 - prBG) / 255.0f;
            f3 *= rho + g * prFG;
            //f3 = 1.0f; /// OJO QUE HEMOS ANULADO EL FACTOR 3
//            alternatef3Sum += g * prFG;
//            alternatef3Weights += g;
        }
    }

    if (0 == j) {
        printf("j 0: Particle %u: associated label %u\n", partIdx, thisParticle.associatedLabel);
    }

    const float precision = static_cast<float>(j_in) / (static_cast<float>(M_PI) * thisParticle.radiusX * thisParticle.radiusY);
    const float recall    = static_cast<float>(j_in) / static_cast<float>(j);

    const float fNumerator   = 2.0f * precision * recall;
    const float fDenominator = precision + recall;
    const float L = (0.0f < fDenominator) ? (fNumerator / fDenominator) : 0.0;

//    const float L = ((likelihoodBeta * precision) + ((1.0f - likelihoodBeta) * recall)) * f3;
////                    * alternatef3Sum / alternatef3Weights;

//    if (L == INFINITY) {
//        printf("L infinito: Particle %u: associated label %u, f3 %f, j_in %u, j %u, rX %f, rY %f\n", partIdx, thisParticle.associatedLabel, f3, j_in, j, thisParticle.radiusX, thisParticle.radiusY);
//    }
//    if (L == 0.0) {
//        printf("L cero: Particle %u: associated label %u, f3 %f, j_in %u, j %u, rX %f, rY %f\n", partIdx, thisParticle.associatedLabel, f3, j_in, j, thisParticle.radiusX, thisParticle.radiusY);
//    }
//    if (L != L) {
//        printf("L NAN: Particle %u: associated label %u, f3 %f, j_in %u, j %u, rX %f, rY %f\n", partIdx, thisParticle.associatedLabel, f3, j_in, j, thisParticle.radiusX, thisParticle.radiusY);
//    }

    // Saturamos en 10^30 (el máximo de un float es ~10^38) para lograr estabilidad numérica
    thisParticle.weight *= (1E30 < L) ? 1E30 : L;
}










void launchComputeParticleCoverageMask (const ImageDimensions    & imageDim,
                                        const ParticleManager    & particles,
                                        bool                     * d_coverageMask,
                                        const dim3               & gridSize,
                                        const dim3               & blockSize,
                                        cudaStream_t               stream) {
    computeParticleCoverageMask<<<gridSize, blockSize, 0, stream>>>(imageDim,
                                                                    particles.host.size(),
                                                                    particles.device,
                                                                    d_coverageMask);
/// The original region coverage criterion was that it was enough that at least a single particle covered at least one pixel.
/// However, this can (and does) lead to the situation where the SV assigned to a region does not have any particle
/// that covers it (the region was covered by a particle from other SV), and this causes pain when weighing the particles.
}

void launchComputeParticleCoverageParity (const ImageDimensions    & imageDim,
                                          const ParticleManager    & particles,
                                          unsigned char            * d_coverageParity,
                                          const dim3               & gridSize,
                                          const dim3               & blockSize,
                                          cudaStream_t               stream) {
    computeParticleCoverageParity<<<gridSize, blockSize, 0, stream>>>(imageDim,
                                                                                                 particles.host.size(),
                                                                                                 particles.device,
                                                                                                 d_coverageParity);
}

void launchDetermineCoveredRegions (      unsigned int         numMaxRegions,
                                    const ImageDimensions    & imageDim,
                                    const PixelInRegion      * d_labelledPixels,
                                          unsigned int         numberOfLabelledPixels,
                                    const bool               * d_coverageMask,
                                          unsigned int       * d_coveredRegions,
                                          cudaStream_t         stream) {
    determineCoveredRegions<<<ceilUintDivision(numberOfLabelledPixels, 256), 256, sizeof(unsigned int) * (numMaxRegions + 1), stream>>>(numMaxRegions, imageDim, d_labelledPixels, numberOfLabelledPixels, d_coverageMask, d_coveredRegions);
}


void launchWeighParticles(const ImageDimensions         & imageDim,
                                float                     likelihoodBeta,
                          const unsigned char           * bgProbability,
                          const PixelInRegion           * d_labelledPixels,
                                unsigned int              numberOfLabelledPixels,
                                ParticleManager         & particles,
                                cudaStream_t              stream) {

    weighParticles<<<ceilUintDivision(particles.host.size(), 256), 256, 0, stream>>>(imageDim, likelihoodBeta, bgProbability, d_labelledPixels, numberOfLabelledPixels, particles.device, particles.host.size());
}


void launchComputeParticleFGPriorProbability(const ImageDimensions & imageDim,
                                             const ParticleManager & particles,
                                                   float           * d_probability,
                                             const dim3            & gridSize,
                                             const dim3            & blockSize,
                                                   cudaStream_t      stream) {
    computeParticleFGPriorProbability<<<gridSize, blockSize, 0, stream>>>(imageDim,
                                                                   particles.host.size(),
                                                                   particles.device,
                                                                   d_probability);
}

void launchComputeConstantFGPriorProbability(const ImageDimensions & imageDim,
                                                   float           * d_probability,
                                             const dim3            & gridSize,
                                             const dim3            & blockSize,
                                                   cudaStream_t      stream) {
    computeConstantFGPriorProbability<<<gridSize, blockSize, 0, stream>>>(imageDim,
                                                                   d_probability);
}
