﻿#include <QtCore/QDebug>
#include <QtCore/QTime>
#include <FreeImagePlus.h>
#include <cuda_runtime.h>
#include <cassert>
#include <cmath>
#include <cstdio>
#include "processingstage.h"
#include "auxcuda.h"
#include "processingstagecuda.h"
#include "processingdatatypes.h"
#include "thumbnail.h"

#include "label.h"

template <typename ELEMENT_TYPE>
ProcessingStage<ELEMENT_TYPE>::ProcessingStage()
    : tracker(1) // Needed to compile because CBT does not accept default
                 // constructor. Impossible to invoke though.
{
}

void setupSpatialGaussian (float        **where,
                           float          sigma,
                           unsigned int   radius,
                           unsigned int   normalisationDimension) {
    assert (0.0f != sigma);




    CUDA_SAFE_CALL (cudaMalloc (where, sizeof (float) * (1 + radius)));
    float *hostGaussian = new float[1 + radius];

    // If radius is UINT_MAX, 1 + radius equals zero, effectively terminating
    // the loop, whereas the (delta <= radius) would be an infinite loop.
    for (unsigned int delta = 0; delta < (1 + radius); ++delta) {
        hostGaussian[delta] = static_cast <float> (exp (-0.5 * static_cast <double> (delta * delta)
                                                        / (sigma * sigma))
                                                   / (sqrt (2.0 * 3.141592654) * sigma
                                                      / static_cast <double> (normalisationDimension)) );
//        qDebug () << hostGaussian[delta];
    }
    CUDA_SAFE_CALL (cudaMemcpy (*where, hostGaussian, sizeof (float) * (1 + radius), cudaMemcpyHostToDevice));
    delete [] hostGaussian;
}

template <typename ELEMENT_TYPE>
void ProcessingStage<ELEMENT_TYPE>::setupSpatialGaussians()
{
    setupSpatialGaussian (&this->deviceAuxData.backgroundSpatialHorizontalGaussian,
                          this->prConfig.backgroundSpatialSigma,
                          this->prConfig.backgroundSpatialRadius,
                          this->imageDim.width);
    setupSpatialGaussian (&this->deviceAuxData.backgroundSpatialVerticalGaussian,
                          this->prConfig.backgroundSpatialSigma,
                          this->prConfig.backgroundSpatialRadius,
                          this->imageDim.height);
    setupSpatialGaussian (&this->deviceAuxData.foregroundSpatialHorizontalGaussian,
                          this->prConfig.foregroundSpatialSigma,
                          this->prConfig.foregroundSpatialRadius,
                          this->imageDim.width);
    setupSpatialGaussian (&this->deviceAuxData.foregroundSpatialVerticalGaussian,
                          this->prConfig.foregroundSpatialSigma,
                          this->prConfig.foregroundSpatialRadius,
                          this->imageDim.height);
}





template <typename ELEMENT_TYPE>
ProcessingStage<ELEMENT_TYPE>::ProcessingStage(const ProcessingConfig & config, Stage *next)
    : prConfig(config),
      tracker(config.referenceBufferSize,
              config.backgroundReferenceFrames) // TODO : What if FG requires more?
{
    this->next = next;
    this->inputImageArray                     = NULL;
    this->imageDiffArray                      = NULL;
//    this->probArray                           = NULL;
//    this->probDiffArray                       = NULL;
//    this->imageDiffSumArray                   = NULL;
//    this->probDiffSumArray                    = NULL;

}

template <typename ELEMENT_TYPE>
ProcessingStage<ELEMENT_TYPE>::~ProcessingStage()
{
    CUDA_SAFE_CALL(cudaFreeArray (this->inputImageArray));
    CUDA_SAFE_CALL(cudaFreeArray (this->imageDiffArray));
//    CUDA_SAFE_CALL(cudaFreeArray (this->probArray));
//    CUDA_SAFE_CALL(cudaFreeArray (this->probDiffArray));
//    CUDA_SAFE_CALL(cudaFreeArray (this->imageDiffSumArray));
//    CUDA_SAFE_CALL(cudaFreeArray (this->probDiffSumArray));





    qDebug ("Salimos del destructor de Processing Stage");
}

template <typename ELEMENT_TYPE>
std::ptrdiff_t ProcessingStage<ELEMENT_TYPE>::startWrite()
{
    return this->tracker.startWrite () * this->imageDim.channels;
}

template <typename ELEMENT_TYPE>
void ProcessingStage<ELEMENT_TYPE>::endWrite(std::ptrdiff_t written)
{
    return this->tracker.endWrite (written / this->imageDim.channels);
}

template <typename ELEMENT_TYPE>
void ProcessingStage<ELEMENT_TYPE>::notifyEndOfStream()
{
    this->tracker.notifyEndOfStream ();
}

template <typename ELEMENT_TYPE>
void ProcessingStage<ELEMENT_TYPE>::run()
{
    QString filename ("");
//    fipImage image;

    ProcessingHostData hostData;
    CUDA_SAFE_CALL (cudaStreamCreate(&hostData.stream));

//    cudaMemcpy3DParms memcpyParams;
//    memset(&memcpyParams, 0, sizeof (memcpyParams));
//    memcpyParams.kind   = cudaMemcpyDeviceToHost;

    QTime timer;
    bool timerStarted = false;
    unsigned int stationaryFrames = 0;

    bool buffersAreInitialized = false;

    for (int i = 0; true ; ++i) {
        filename.sprintf ("/run/shm/labels/labelled%04d.tif", i);
//        qDebug ("Consumer starting read %06d", i);
        hostData.readIndex = this->tracker.startRead();

        if (!buffersAreInitialized) {
            buffersAreInitialized = true;
            hostData.blockSize = dim3(this->prConfig.blockWidth, this->prConfig.blockHeight);
            hostData.gridSize  = dim3 (1 + (this->imageDim.width  - 1) / hostData.blockSize.x,
                                       1 + (this->imageDim.height - 1) / hostData.blockSize.y);
            hostData.thumbnail.allocate(hostData.gridSize.x, hostData.gridSize.y);
            hostData.particleFilter.initialize(this->prConfig, this->imageDim, hostData.gridSize, hostData.blockSize, hostData.stream, false);
            hostData.thumbnail.setRandomSeed(666);
            hostData.thumbnail.setEnabled(this->prConfig.roiAnalysis);
        }

        if (-1 == hostData.readIndex) {
//            qDebug ("Esto se ha acabado, chavales!");
//            qDebug () << "Grid dimensions" << hostData.gridSize.x << " x " << hostData.gridSize.y;
            break;
        }



        hostData.writeIndex = this->next->startWrite();

        hostData.availableReference = this->tracker.getAvailablePast();



        if (hostData.availableReference >= static_cast<unsigned int>(this->prConfig.backgroundReferenceFrames)) {
            if (!timerStarted) {
                timerStarted = true;
                timer.start ();
            }
            stationaryFrames++;
        }

        launchProcess<ELEMENT_TYPE> (this->prConfig,
                                      this->imageDim,
                                      this->deviceAuxData,
                                      this->deviceData,
                                      hostData);
        this->next->endWrite (hostData.writeIndex);

//        qDebug ("Consumer ending read %06d", i);
        this->tracker.endRead(hostData.readIndex);
//        assert (image.save (filename.toStdString ().c_str ()));
    }
    int elapsed = timer.elapsed ();
    qDebug () << "Stationary processing time (ms)" << elapsed;
    qDebug() << "Mean time per frame (ms):" << (static_cast<float>(elapsed) / static_cast<float>(stationaryFrames));
    qDebug() << "Mean time per pixel (ns)" << 1000000.0f * (static_cast<float>(elapsed) / static_cast<float>(stationaryFrames * this->imageDim.height * this->imageDim.width));
}

template <typename ELEMENT_TYPE>
cudaArray * ProcessingStage<ELEMENT_TYPE>::initializeBuffer(unsigned int width,
                                             unsigned int height,
                                             unsigned int channels)
{
    this->imageDim.width = width;
    this->imageDim.height = height;
    this->imageDim.channels = channels;
    return this->initializeBuffer ();
}

template <typename ELEMENT_TYPE>
cudaArray * ProcessingStage<ELEMENT_TYPE>::initializeBuffer()
{

    //assert (this->imageDim.channels <= MAX_CHANNELS); // TODO : Equality is better

    if ((this->prConfig.backgroundSpatialModelling) || (this->prConfig.foregroundModelEnabled)) {
        this->setupSpatialGaussians ();
    }

    cudaChannelFormatDesc const inputFormat = makeCudaChannelFormatDesc<ELEMENT_TYPE>();
    CUDA_SAFE_CALL (cudaMalloc3DArray (&this->inputImageArray,
                                       &inputFormat,
                                       make_cudaExtent(this->imageDim.width,
                                                       this->imageDim.height,
                                                       this->imageDim.channels * this->tracker.getCapacity ()),
                                       cudaArrayLayered | cudaArraySurfaceLoadStore));
    CUDA_SAFE_CALL (cudaMalloc3DArray (&this->imageDiffArray,
                                       &inputFormat,
                                       make_cudaExtent(this->imageDim.width,
                                                       this->imageDim.height,
                                                       this->imageDim.channels * this->tracker.getCapacity ()),
                                       cudaArrayLayered | cudaArraySurfaceLoadStore));

    this->deviceData.allocateDeviceMemory(this->imageDim.width,
                                          this->imageDim.height,
                                          this->imageDim.channels,
                                          this->tracker.getCapacity(),
                                          this->prConfig.foregroundReferenceFrames);

    this->outputImageArray = this->next->initializeBuffer (this->imageDim.width, this->imageDim.height, 1);
    assert (NULL != this->outputImageArray);
    bindProcessingSurfaces (this->inputImageArray,
                            this->outputImageArray
                            );
    return this->inputImageArray;
}

template <typename ELEMENT_TYPE>
cudaArray * ProcessingStage<ELEMENT_TYPE>::initializeBuffer(const ImageDimensions &dim)
{
    this->imageDim = dim;
    return this->initializeBuffer ();
}

// Explicit instantiation
template class ProcessingStage<unsigned char>;
template class ProcessingStage<unsigned short>;
