#include <cfloat>
#include <cassert>
#include <thrust/device_ptr.h>
#include <thrust/reduce.h>
#include <thrust/transform_reduce.h>

#include <cstring>
#include <cstdio>
#include <cmath>
#include <FreeImagePlus.h>
#include "label.h"

#include <set>
#include <algorithm>
#include <vector>

#include "auxcuda.h"
#include "processingstagecuda.h"
#include "scalarutils.h"
#include "imagedimensions.h"
#include "processingconfig.h"
#include "snd.h"
#include "thumbnail.h"
#include "particleset.h"
#include "statevectorsset.h"
#include "prettyprintstl.h"
#include "particlefilter.h"
#include "linearforegroundmodel.h"
#include "compiletimeconfig.h"

unsigned int const WARP_SIZE = 32;

surface<void, cudaSurfaceType2DLayered> inputImageSurface;
surface<void, cudaSurfaceType2DLayered> nextStageSurface;




texture<float, cudaTextureType1D, cudaReadModeElementType> sndOrdinatesTex;
static cudaArray *sndOrdinatesArray;
texture<float, cudaTextureType1D, cudaReadModeElementType> expOrdinatesTex;
static cudaArray *expOrdinatesArray;



void setupUniformGaussian () {
    const cudaChannelFormatDesc formatFloat = {32, 0, 0, 0, cudaChannelFormatKindFloat};
    CUDA_SAFE_CALL (cudaMalloc3DArray (&sndOrdinatesArray,
                                          &formatFloat,
                                          make_cudaExtent(sndElements, 0, 0),
                                          cudaArrayDefault));
    cudaMemcpy3DParms memcpyParams = {0};
    memcpyParams.extent = make_cudaExtent (sndElements, 1, 1); // Docs say width should be bytes, but it's actually elements
    memcpyParams.dstPos = make_cudaPos (0, 0, 0);
    memcpyParams.kind   = cudaMemcpyHostToDevice;
    memcpyParams.srcPtr = make_cudaPitchedPtr (const_cast<float *> (sndValues),
                                               sizeof (float) * sndElements,
                                               sndElements,
                                               1);
    memcpyParams.dstArray = sndOrdinatesArray;
    CUDA_SAFE_CALL (cudaMemcpy3D (&memcpyParams));

    // Bind the array containing the SND ordinates to the corresponding 1-D texture
    sndOrdinatesTex.addressMode[0] = cudaAddressModeBorder; // Returns 0 outside bounds
    sndOrdinatesTex.filterMode = cudaFilterModeLinear; // Linear interpolation
    sndOrdinatesTex.normalized = false; // Coordinates are still from 0 to points - 1
    CUDA_SAFE_CALL(cudaBindTextureToArray(sndOrdinatesTex, sndOrdinatesArray));
}


__device__ __forceinline__ float standardNormal (float x) {
    float const absXInPitches = ((x < 0.0f) ? -x : x) * sndInvPitch;// INV_UNIF_PITCH;
    return tex1D(sndOrdinatesTex, absXInPitches + 0.5f);
}


void setupNegativeExponential () {
    const cudaChannelFormatDesc formatFloat = {32, 0, 0, 0, cudaChannelFormatKindFloat};
    CUDA_SAFE_CALL (cudaMalloc3DArray (&expOrdinatesArray,
                                          &formatFloat,
                                          make_cudaExtent(expElements, 0, 0),
                                          cudaArrayDefault));
    cudaMemcpy3DParms memcpyParams = {0};
    memcpyParams.extent = make_cudaExtent (expElements, 1, 1); // Docs say width should be bytes, but it's actually elements
    memcpyParams.dstPos = make_cudaPos (0, 0, 0);
    memcpyParams.kind   = cudaMemcpyHostToDevice;
    memcpyParams.srcPtr = make_cudaPitchedPtr (const_cast<float *> (expValues),
                                               sizeof (float) * expElements,
                                               expElements,
                                               1);
    memcpyParams.dstArray = expOrdinatesArray;
    CUDA_SAFE_CALL (cudaMemcpy3D (&memcpyParams));

    // Bind the array containing the SND ordinates to the corresponding 1-D texture
    expOrdinatesTex.addressMode[0] = cudaAddressModeBorder; // Returns 0 outside bounds
    expOrdinatesTex.filterMode = cudaFilterModeLinear; // Linear interpolation
    expOrdinatesTex.normalized = false; // Coordinates are still from 0 to points - 1
    CUDA_SAFE_CALL(cudaBindTextureToArray(expOrdinatesTex, expOrdinatesArray));
}


__device__ __forceinline__ float negExp (float x) {
    return tex1D(expOrdinatesTex, x * expInvPitch + 0.5f);
}


void bindProcessingSurfaces (cudaArray *inputImageArray,
                             cudaArray *nextStageArray
                             ) {
    CUDA_SAFE_CALL (cudaBindSurfaceToArray (inputImageSurface, inputImageArray));
    CUDA_SAFE_CALL (cudaBindSurfaceToArray (nextStageSurface, nextStageArray));
    setupUniformGaussian ();
    setupNegativeExponential ();
}




// These templates generate instructions in compile time for computing
// the product of all values of a vector "sequence" of length "N" without
// having to use a loop. Since N needs to be known in compile time, a good
// compiler would unroll such loop anyway, but this is way cooler! :-P
template <unsigned int N>
__device__ float computeProduct (float sequence[]) {
    return sequence[N-1] * computeProduct<N-1> (sequence);
}

template <>
__device__ float computeProduct<0> (float /* sequence */ []) {
    return 1.0f;
}

template <unsigned int EXP>
__device__ float computePower (float base) {
    return base * computePower<EXP-1> (base);
}

template <>
__device__ float computePower<0> (float base) {
    return 1.0f;
}




//template <typename T>
//__forceinline__ __device__ __host__ T* displacePtr3D (T *basePtr, int xSize, int ySize, int x, int y, int z) {
//    return basePtr
//           + x
//           + y * xSize
//           + z * xSize * ySize;
//}

//template <typename T>
//__forceinline__ __device__ __host__ T const* displacePtr3D (T const *basePtr, int xSize, int ySize, int x, int y, int z) {
//    return basePtr
//           + x
//           + y * xSize
//           + z * xSize * ySize;
//}

float __device__ clamp(float v, float minV, float maxV) {
    return fminf(maxV, fmaxf(minV, v));
}



/// We play at that max is set high enough that the fraction of samples that exceeds it is
/// negligible. If that is not the case, our estimate will be seriously wrong
__device__ unsigned char float2bin (float v) {
    const float logBins = NUMBINS - 1;
    if (v < MINDIFF) {
        return 0;
    } else if (v >= MAXDIFF) {
        return NUMBINS - 1;
    } else {
        //return 1 + static_cast<unsigned int>(logBins * (__logf(v) - __logf(MINDIFF)) / (__logf(MAXDIFF) - __logf(MINDIFF)));
        return 1 + static_cast<unsigned int>(logBins * __fdividef(__logf(v) - __logf(MINDIFF), __logf(MAXDIFF) - __logf(MINDIFF)));
    }
}

/// For now, we just return the value corresponding to the middle of the interval in log scale
__device__ float bin2float (unsigned char bin) {
    const float logBins = NUMBINS - 1;
    if (0 == bin) {
        return MINDIFF;
    } else {
        const float logMinDiff = logf(MINDIFF); // Si pongo __logf... ¡¡¡peta el compilador!!!
        const float logMaxDiff = __logf(MAXDIFF);
        return __expf(logMinDiff + ((bin - 0.5f) / logBins * (logMaxDiff - logMinDiff)));
        //return expf(__logf(MINDIFF) + ((bin - 0.5f) / logBins * (__logf(MAXDIFF) - __logf(MINDIFF))));
        //return __expf(__logf(MINDIFF) + (__fdividef(bin - 0.5f, logBins) * (__logf(MAXDIFF) - __logf(MINDIFF))));
    }
}







template <typename ELEMENT_TYPE,
          unsigned int CHANNELS>
__device__ float getAppearanceSigma (ProcessingConfig  config,
                                     ImageDimensions   imageDim,
                                     float const *__restrict__ backgroundSpatialHorizontalGaussian,
                                     float const *__restrict__ backgroundSpatialVerticalGaussian,
                                     long const  *__restrict__ imageSquaredDiffTemporalWeightedSum,
                                     long const  *__restrict__ probTemporalSum,
                                     unsigned int channel) {
    //if (channel >= MAX_CHANNELS) { asm ("trap;"); }

    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;

    float sigma = 0.0f;
    float normalisationFactor = 0.0f;
    float const sqrt2 = 1.414213562373095f;

    // Multiplying radiuses by BSP tells the compiler that the loop is not really a loop when it is disabled
    for (int deltaY = -config.backgroundSpatialRadius;
             deltaY <= config.backgroundSpatialRadius;
         ++deltaY) {
        for (int deltaX = -config.backgroundSpatialRadius;
                 deltaX <= config.backgroundSpatialRadius;
             ++deltaX) {
            float const spatialWeight = config.backgroundSpatialModelling
                                        ? backgroundSpatialHorizontalGaussian[abs(deltaX)] * backgroundSpatialVerticalGaussian[abs(deltaY)]
                                        : 1.0f; // Neutral element if we are not using spatial modelling
            int const readX = x + deltaX;
            int const readY = y + deltaY;
            if (   (readX >= 0)                && (readY >= 0)
                && (readX < imageDim.width) && (readY < imageDim.height)) {
                float const probSum = static_cast<float> (*(displacePtr3D<long>(probTemporalSum,
                                                                                imageDim.width,
                                                                                imageDim.height,
                                                                                readX, readY,
                                                                                0)));
                if (0.0f == probSum) {
                    continue; // We do now want to divide by zero
                }
                normalisationFactor += spatialWeight;
                float const appearanceSum = static_cast<float> (*(displacePtr3D<long>(imageSquaredDiffTemporalWeightedSum,
                                                                                      imageDim.width,
                                                                                      imageDim.height,
                                                                                      readX, readY,
                                                                                      channel)));
                sigma += spatialWeight * sqrtf (appearanceSum / probSum);
            }
        }
    }
    if (0.0f == normalisationFactor) {
        return config.backgroundAppearanceSigmaEstimationMinimum;
    } else {
        normalisationFactor *= (sqrt2 * ScalarInfo<ELEMENT_TYPE>::NORM_MAX);
        sigma /= normalisationFactor;
    }

    return clamp(sigma, config.backgroundAppearanceSigmaEstimationMinimum, config.backgroundAppearanceSigmaEstimationMaximum);

//    if (sigma < config.backgroundAppearanceSigmaEstimationMinimum) {
//        return config.backgroundAppearanceSigmaEstimationMinimum;
//    } else if (sigma > config.backgroundAppearanceSigmaEstimationMaximum) {
//        return config.backgroundAppearanceSigmaEstimationMaximum;
//    } else {
//        return sigma;
//    }
}






__device__ inline float computeG_T (ProcessingConfig const &config,
                                    int deltaN_i) {
    const int diff = deltaN_i - config.backgroundSelectiveUpdateCutoff;
    return expf (-static_cast<float>(diff * diff)
                 / (config.backgroundSelectiveUpdateSigmaTsquare * 2.0f));
}

__global__ void uniformForegroundModelKernel (ProcessingConfig                  config,
                                              ImageDimensions                   imageDim,
                                              float              * __restrict__ pdfFG,
                                              unsigned char      * __restrict__ thumbnail) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    if (config.roiAnalysis) {
        switch (thumbnail[blockIdx.x + blockIdx.y * gridDim.x]) {
        case Thumbnail::NOT_ANALYSED:
            // fallthrough
        case Thumbnail::POSITIVE:
            ;// fallthrough
        case Thumbnail::NEGATIVE:
            return;
        }
    }

    *(displacePtr3D<float>(pdfFG,
                           imageDim.width,
                           imageDim.height,
                           x, y, 0)) = 1.0f;
}



/// Esto es el conato de modelo de FG con motion vectors, pero al final va a ser más práctico hacerlo
/// poniendo los datos de otra forma.
//template <typename ELEMENT_TYPE,
//          unsigned int CHANNELS>
//__global__ void foregroundModelKernel (ProcessingConfig                  config,
//                                       ImageDimensions                   imageDim,
//                                       int                               availablePastFrames,
//                                       int                               inputIndex,
//                                       float const        * __restrict__ foregroundSpatialHorizontalGaussian,
//                                       float const        * __restrict__ foregroundSpatialVerticalGaussian,
//                                       int   const        * __restrict__ numberOfFGSamples,
//                                       short2 const       * __restrict__ motionVectors,
//                                       unsigned char      * __restrict__ bgProbabilities,
//                                       float              * __restrict__ pdfFG,
//                                       unsigned char      * __restrict__ thumbnail) {
//    int const outputX = blockIdx.x * blockDim.x + threadIdx.x;
//    int const outputY = blockIdx.y * blockDim.y + threadIdx.y;
//    if ((0 > outputX) || (0 > outputY) || (outputX >= imageDim.width) || (outputY >= imageDim.height)) { return; }

//    if (config.roiAnalysis) {
//        switch (thumbnail[blockIdx.x + blockIdx.y * gridDim.x]) {
//        case Thumbnail::NOT_ANALYSED:
//            // fallthrough
//        case Thumbnail::POSITIVE:
//            ;// fallthrough
//        case Thumbnail::NEGATIVE:
//            return;
//        }
//    }

//    // Read current samples
//    int currentImageSamples[CHANNELS];
//    #pragma unroll
//    for (int ch = 0; ch < CHANNELS; ++ch) {
//    //TODO: Si por alguna razón quisiéramos usar floats esto estaría mal
//        currentImageSamples[ch] = static_cast<int>(surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * outputX, outputY, inputIndex * CHANNELS + ch));
//    }

//    // Compute the number of available reference frames
//    int const referenceFrames = min(availablePastFrames, config.foregroundReferenceFrames);
//    float const invAppearanceSigma = -0.5 / (static_cast<double>(ScalarInfo<ELEMENT_TYPE>::NORM_MAX)
//                                             * static_cast<double>(ScalarInfo<ELEMENT_TYPE>::NORM_MAX)
//                                             * static_cast<double>(config.foregroundAppearanceSigma)
//                                             * static_cast<double>(config.foregroundAppearanceSigma)
//                                             );

//    float sumAccumulatorForFG = 0.0f;
//    float marginalSumAccumulatorForFG = 0.0f;
//    float weightAccumulatorForFG = 0.0f;

//    int referenceX = outputX;
//    int referenceY = outputY;


//    for (int deltaT = 1; deltaT <= referenceFrames; ++deltaT) {
//        /// referenceFrames < referenceBufferSize, hence:
//        /// abs(inputIndex - deltaT) < config.referenceBufferSize ==> referenceIndex >= 0
//        int const referenceIndex = ((inputIndex - deltaT) < 0)
//                                   ? (inputIndex - deltaT + config.referenceBufferSize)
//                                   : (inputIndex - deltaT);

//        if (config.foregroundSpatialModelling) {
//            weightAccumulatorForFG += static_cast<float> (numberOfFGSamples[referenceIndex]);
//        }

//        // In each frame, the particle filter estimates a motion for each region. When
//        const short2 & motionVector = *displacePtr3D(motionVectors, imageDim.width, imageDim.height, referenceX, referenceY, referenceIndex);
//        referenceX += motionVector.x;
//        referenceY += motionVector.y;


//        /// TODO : Por temas de coalescing y tal es más que probable que sea mejor, dado que ahora
//        /// tenemos las imágenes como planos separados, que el bucle más exterior sea el de canales
//        /// más que el espacial. Además sería bueno deshacerse del rollo de tener que usar
//        /// arrayes para guardar la muestra original.
//        /// Lamentablemente, el bucle interior es un productorio mientras que los exteriores
//        /// son sumatorios. Además, mover los canales para fuera nos obligaría a leer la
//        /// probabilidad una vez por canal en el caso de la actualización selectiva.
//        for (int deltaY = -config.foregroundSpatialRadius;
//                 deltaY <= config.foregroundSpatialRadius;
//             ++deltaY) {
//            for (int deltaX = -config.foregroundSpatialRadius;
//                     deltaX <= config.foregroundSpatialRadius;
//                 ++deltaX) {
//                int const efX = referenceX + deltaX;
//                int const efY = referenceY + deltaY;
//                if (   (efX >= 0)             && (efY >= 0)
//                    && (efX < imageDim.width) && (efY < imageDim.height)) {

//                    float const referenceProbBG = static_cast<float> (*(displacePtr3D<unsigned char>(bgProbabilities,
//                                                                                                     imageDim.width,
//                                                                                                     imageDim.height,
//                                                                                                     efX, efY, referenceIndex)))
//                                                  / static_cast<float> (UCHAR_MAX);

//                    // We only consider the sample if it was classified as FG
//                    if (0.5f > referenceProbBG) {
//                        if (!config.foregroundSpatialModelling) {
//                            weightAccumulatorForFG++;
//                        }
//                        float const spatialProduct = config.foregroundSpatialModelling
//                                                     ? foregroundSpatialHorizontalGaussian[abs(deltaX)] * foregroundSpatialVerticalGaussian[abs(deltaY)]
//                                                     : 1.0f; // Neutral element if we are not using spatial modelling
//                        float partialProduct  = spatialProduct;
//                        float const marginalProduct = spatialProduct;
//                        float exponent = 0.0f;
//                        #pragma unroll
//                        for (int ch = 0; ch < CHANNELS; ++ch) {
//                            //TODO: Si por alguna razón quisiéramos usar floats esto estaría mal
//                            int const referenceImageSample = static_cast<int> (surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * efX, efY, referenceIndex * CHANNELS + ch));
//                            exponent += static_cast<long>(referenceImageSample - currentImageSamples[ch])
//                                        * static_cast<long>(referenceImageSample - currentImageSamples[ch])
//                                        * invAppearanceSigma;
//                        }
//                        partialProduct *= __expf (exponent) * InvSqrt2Pi<CHANNELS>::VALUE;
//                        //partialProduct *= negExp (-exponent);
//                        sumAccumulatorForFG += partialProduct;
//                        marginalSumAccumulatorForFG += marginalProduct;
//                    }
//                }
//            }
//        }
//    }

//    float const & alpha = config.foregroundMixingFactor;
//    float const & sigma = config.foregroundAppearanceSigma;

//    const float pdfForFG = (0.0f == weightAccumulatorForFG)
//                           ? 1.0f
//                           : (alpha + (1.0f - alpha) * sumAccumulatorForFG / (computePower<CHANNELS> (sigma) * weightAccumulatorForFG))
//                             / (config.spatialMarginalConditioning
//                                ? (alpha + (1.0f - alpha) * marginalSumAccumulatorForFG /  weightAccumulatorForFG)
//                                : 1.0f);



//    *(displacePtr3D<float>(pdfFG,
//                           imageDim.width,
//                           imageDim.height,
//                           outputX, outputY, 0)) = pdfForFG;
//}

//template <typename ELEMENT_TYPE,
//          unsigned int CHANNELS>
//__global__ void linearForegroundModelKernel (ProcessingConfig                  config,
//                                             ImageDimensions                   imageDim,
//                                             int                               availablePastFrames,
//                                             int                               inputIndex,
//                                             float const        * __restrict__ foregroundSpatialHorizontalGaussian,
//                                             float const        * __restrict__ foregroundSpatialVerticalGaussian,
//                                             int   const        * __restrict__ numberOfFGSamples,
//                                             ColouredPixel<ELEMENT_TYPE, CHANNELS> * __restrict__ foregroundPixels,
//                                             unsigned int                      foregroundPixelsLength,
//                                             float              * __restrict__ pdfFG,
//                                             unsigned char      * __restrict__ thumbnail) {
//    int const x = blockIdx.x * blockDim.x + threadIdx.x;
//    int const y = blockIdx.y * blockDim.y + threadIdx.y;
//    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

//    if (config.roiAnalysis) {
//        switch (thumbnail[blockIdx.x + blockIdx.y * gridDim.x]) {
//        case Thumbnail::NOT_ANALYSED:
//            // fallthrough
//        case Thumbnail::POSITIVE:
//            ;// fallthrough
//        case Thumbnail::NEGATIVE:
//            return;
//        }
//    }

//    // Read current samples
//    int currentImageSamples[CHANNELS];
//    #pragma unroll
//    for (int ch = 0; ch < CHANNELS; ++ch) {
//    //TODO: Si por alguna razón quisiéramos usar floats esto estaría mal
//        currentImageSamples[ch] = static_cast<int>(surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * x, y, inputIndex * CHANNELS + ch));
//    }

//    // Compute the number of available reference frames
//    int const referenceFrames = min(availablePastFrames, config.foregroundReferenceFrames);
//    float const invAppearanceSigma = -0.5 / (static_cast<double>(ScalarInfo<ELEMENT_TYPE>::NORM_MAX)
//                                             * static_cast<double>(ScalarInfo<ELEMENT_TYPE>::NORM_MAX)
//                                             * static_cast<double>(config.foregroundAppearanceSigma)
//                                             * static_cast<double>(config.foregroundAppearanceSigma)
//                                             );

//    float sumAccumulatorForFG = 0.0f;
//    float marginalSumAccumulatorForFG = 0.0f;
//    float weightAccumulatorForFG = 0.0f;

//    for (int deltaT = 1; deltaT <= referenceFrames; ++deltaT) {
//        /// referenceFrames < referenceBufferSize, hence:
//        /// abs(inputIndex - deltaT) < config.referenceBufferSize ==> referenceIndex >= 0
//        int const referenceIndex = ((inputIndex - deltaT) < 0)
//                                   ? (inputIndex - deltaT + config.referenceBufferSize)
//                                   : (inputIndex - deltaT);

//        if (config.foregroundSpatialModelling) {
//            weightAccumulatorForFG += static_cast<float> (numberOfFGSamples[referenceIndex]);
//        }
//    }
//    for (unsigned int pxi = 0; pxi < foregroundPixelsLength; ++pxi) {
//        const ColouredPixel<ELEMENT_TYPE, CHANNELS> & px = foregroundPixels[pxi];
//        const int deltaX = abs(x - static_cast<int>(px.col));
//        const int deltaY = abs(y - static_cast<int>(px.row));
//        if ((deltaX <= config.foregroundSpatialRadius) && (deltaY <= config.foregroundSpatialRadius)) {
//            if (!config.foregroundSpatialModelling) {
//                weightAccumulatorForFG++;
//            }
//            float const spatialProduct = config.foregroundSpatialModelling
//                                         ? foregroundSpatialHorizontalGaussian[abs(deltaX)] * foregroundSpatialVerticalGaussian[abs(deltaY)]
//                                         : 1.0f; // Neutral element if we are not using spatial modelling
//            float partialProduct  = spatialProduct;
//            float const marginalProduct = spatialProduct;
//            float exponent = 0.0f;
//            #pragma unroll
//            for (int ch = 0; ch < CHANNELS; ++ch) {
//                //TODO: Si por alguna razón quisiéramos usar floats esto estaría mal
//                int const referenceImageSample = static_cast<int> (px.colour[ch]);
//                exponent += static_cast<long>(referenceImageSample - currentImageSamples[ch])
//                            * static_cast<long>(referenceImageSample - currentImageSamples[ch])
//                            * invAppearanceSigma;
//            }
//            partialProduct *= __expf (exponent) * InvSqrt2Pi<CHANNELS>::VALUE;
//            //partialProduct *= negExp (-exponent);
//            sumAccumulatorForFG += partialProduct;
//            marginalSumAccumulatorForFG += marginalProduct;
//        }
//    }

//    float const & alpha = config.foregroundMixingFactor;
//    float const & sigma = config.foregroundAppearanceSigma;

//    const float pdfForFG = (0.0f == weightAccumulatorForFG)
//                           ? 1.0f
//                           : (alpha + (1.0f - alpha) * sumAccumulatorForFG / (computePower<CHANNELS> (sigma) * weightAccumulatorForFG))
//                             / (config.spatialMarginalConditioning
//                                ? (alpha + (1.0f - alpha) * marginalSumAccumulatorForFG /  weightAccumulatorForFG)
//                                : 1.0f);



//    *(displacePtr3D<float>(pdfFG,
//                           imageDim.width,
//                           imageDim.height,
//                           x, y, 0)) = pdfForFG;
//}



template <typename ELEMENT_TYPE,
          unsigned int CHANNELS>
__global__ void multiSampleForegroundModelKernel (ProcessingConfig                  config,
                                                  ImageDimensions                   imageDim,
                                                  int                               availablePastFrames,
                                                  int                               inputIndex,
                                                  float const        * __restrict__ foregroundSpatialHorizontalGaussian,
                                                  float const        * __restrict__ foregroundSpatialVerticalGaussian,
                                                  int   const        * __restrict__ numberOfFGSamples,

                                                  ELEMENT_TYPE const * __restrict__ fgReferenceSamples,
                                                  unsigned int const * __restrict__ fgSampleCounts,

                                                  float const        * __restrict__ estimatedBGAppearanceSigmas,
                                                  float              * __restrict__ pdfFG,
                                                  unsigned char const * __restrict__ thumbnail) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    if (config.roiAnalysis) {
        switch (thumbnail[blockIdx.x + blockIdx.y * gridDim.x]) {
        case Thumbnail::NOT_ANALYSED:
            // fallthrough
        case Thumbnail::POSITIVE:
            ;// fallthrough
        case Thumbnail::NEGATIVE:
            return;
        }
    }

    // Read current samples
    int currentImageSamples[CHANNELS];
    #pragma unroll
    for (int ch = 0; ch < CHANNELS; ++ch) {
    //TODO: Si por alguna razón quisiéramos usar floats esto estaría mal
        currentImageSamples[ch] = static_cast<int>(surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * x, y, inputIndex * CHANNELS + ch));
    }

    // Compute the number of available reference frames
    int const referenceFrames = min(availablePastFrames, config.foregroundReferenceFrames);

    float invAppearanceSigmas[CHANNELS];
    float appearanceSigmasProduct = 1.0f;
    {
        const Pointer3D<float> estimatedAppearanceSigmasPtr(const_cast<float *>(estimatedBGAppearanceSigmas), imageDim.width, imageDim.height);
        for (int ch = 0; ch < CHANNELS; ++ch) {
            const float currentAppearanceSigma = (config.backgroundAppearanceSigmaEstimation && config.foregroundUseBGAppearanceSigma)
                                                 ? config.foregroundBGAppearanceSigmaMultiplier * estimatedAppearanceSigmasPtr(x, y, ch)
                                                 : config.foregroundAppearanceSigma;
            appearanceSigmasProduct *= currentAppearanceSigma;

            invAppearanceSigmas[ch] = -0.5 / (static_cast<double>(ScalarInfo<ELEMENT_TYPE>::NORM_MAX)
                                                      * static_cast<double>(ScalarInfo<ELEMENT_TYPE>::NORM_MAX)
                                                      * static_cast<double>(currentAppearanceSigma)
                                                      * static_cast<double>(currentAppearanceSigma)
                                                      );
        }
    }

    float sumAccumulatorForFG = 0.0f;
    float marginalSumAccumulatorForFG = 0.0f;
    float weightAccumulatorForFG = 0.0f;

    for (int deltaT = 1; deltaT <= referenceFrames; ++deltaT) {
        /// referenceFrames < referenceBufferSize, hence:
        /// abs(inputIndex - deltaT) < config.referenceBufferSize ==> referenceIndex >= 0
        int const referenceIndex = ((inputIndex - deltaT) < 0)
                                   ? (inputIndex - deltaT + config.referenceBufferSize)
                                   : (inputIndex - deltaT);

        if (config.foregroundSpatialModelling) {
            weightAccumulatorForFG += static_cast<float> (numberOfFGSamples[referenceIndex]);
        }
    }

    // TODO : FIXME : Mirar si hay alguna solución más estupenda que esto del const_cast
    const Pointer2D<unsigned int> sampleCount2D(const_cast<unsigned int*>(fgSampleCounts), imageDim.width);
    const Pointer4D<ELEMENT_TYPE> samples4D(const_cast<ELEMENT_TYPE *>(fgReferenceSamples), CHANNELS, imageDim.width, imageDim.height);

    for (unsigned int depth = 0; true; ++depth) {
        bool aSampleWasRead = false;



        for (int deltaY = -config.foregroundSpatialRadius;
                 deltaY <= config.foregroundSpatialRadius;
             ++deltaY) {
            for (int deltaX = -config.foregroundSpatialRadius;
                     deltaX <= config.foregroundSpatialRadius;
                 ++deltaX) {
                int const efX = x + deltaX;
                int const efY = y + deltaY;
                if (   (efX >= 0)             && (efY >= 0)
                    && (efX < imageDim.width) && (efY < imageDim.height)) {

                    if (sampleCount2D(efX, efY) > depth) {
//                        printf("asdfasdfads \n");
                        aSampleWasRead = true;
                        if (!config.foregroundSpatialModelling) {
                            weightAccumulatorForFG++;
                        }
                        float const spatialProduct = config.foregroundSpatialModelling
                                                     ? foregroundSpatialHorizontalGaussian[abs(deltaX)] * foregroundSpatialVerticalGaussian[abs(deltaY)]
                                                     : 1.0f; // Neutral element if we are not using spatial modelling
                        float partialProduct  = spatialProduct;
                        float const marginalProduct = spatialProduct;
                        float exponent = 0.0f;
                        #pragma unroll
                        for (int ch = 0; ch < CHANNELS; ++ch) {
                            //TODO: Si por alguna razón quisiéramos usar floats esto estaría mal
                            int const referenceImageSample = static_cast<int> (samples4D(ch, efX, efY, depth));
                            exponent += static_cast<long>(referenceImageSample - currentImageSamples[ch])
                                        * static_cast<long>(referenceImageSample - currentImageSamples[ch])
                                        * invAppearanceSigmas[ch];
                        }
                        partialProduct *= __expf (exponent) * InvSqrt2Pi<CHANNELS>::VALUE;
                        //partialProduct *= negExp (-exponent);
                        sumAccumulatorForFG += partialProduct;
                        marginalSumAccumulatorForFG += marginalProduct;
                    }
                }
            }
        }
        if (!aSampleWasRead) {
            break;
        }
    }





    float const & alpha = config.foregroundMixingFactor;

    const float pdfForFG = (0.0f == weightAccumulatorForFG)
                           ? 1.0f
                           : (alpha + (1.0f - alpha) * sumAccumulatorForFG / (appearanceSigmasProduct * weightAccumulatorForFG))
                             / (config.spatialMarginalConditioning
                                ? (alpha + (1.0f - alpha) * marginalSumAccumulatorForFG /  weightAccumulatorForFG)
                                : 1.0f);



    *(displacePtr3D<float>(pdfFG,
                           imageDim.width,
                           imageDim.height,
                           x, y, 0)) = pdfForFG;
}



template <typename ELEMENT_TYPE,
          unsigned int CHANNELS>
__global__ void foregroundModelKernel (ProcessingConfig                  config,
                                       ImageDimensions                   imageDim,
                                       int                               availablePastFrames,
                                       int                               inputIndex,
                                       float const        * __restrict__ foregroundSpatialHorizontalGaussian,
                                       float const        * __restrict__ foregroundSpatialVerticalGaussian,
                                       int   const        * __restrict__ numberOfFGSamples,
                                       unsigned char const * __restrict__ fgMasks,
                                       float const        * __restrict__ estimatedBGAppearanceSigmas,
                                       float              * __restrict__ pdfFG,
                                       unsigned char const * __restrict__ thumbnail) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    if (config.roiAnalysis) {
        switch (thumbnail[blockIdx.x + blockIdx.y * gridDim.x]) {
        case Thumbnail::NOT_ANALYSED:
            // fallthrough
        case Thumbnail::POSITIVE:
            ;// fallthrough
        case Thumbnail::NEGATIVE:
            return;
        }
    }

    // Read current samples
    int currentImageSamples[CHANNELS];
    #pragma unroll
    for (int ch = 0; ch < CHANNELS; ++ch) {
    //TODO: Si por alguna razón quisiéramos usar floats esto estaría mal
        currentImageSamples[ch] = static_cast<int>(surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * x, y, inputIndex * CHANNELS + ch));
    }

    // Compute the number of available reference frames
    int const referenceFrames = min(availablePastFrames, config.foregroundReferenceFrames);

    float invAppearanceSigmas[CHANNELS];
    float appearanceSigmasProduct = 1.0f;
    {
        const Pointer3D<float> estimatedAppearanceSigmasPtr(const_cast<float *>(estimatedBGAppearanceSigmas), imageDim.width, imageDim.height);
        for (int ch = 0; ch < CHANNELS; ++ch) {
            const float currentAppearanceSigma = (config.backgroundAppearanceSigmaEstimation && config.foregroundUseBGAppearanceSigma)
                                                 ? config.foregroundBGAppearanceSigmaMultiplier * estimatedAppearanceSigmasPtr(x, y, ch)
                                                 : config.foregroundAppearanceSigma;
            appearanceSigmasProduct *= currentAppearanceSigma;

            invAppearanceSigmas[ch] = -0.5 / (static_cast<double>(ScalarInfo<ELEMENT_TYPE>::NORM_MAX)
                                                      * static_cast<double>(ScalarInfo<ELEMENT_TYPE>::NORM_MAX)
                                                      * static_cast<double>(currentAppearanceSigma)
                                                      * static_cast<double>(currentAppearanceSigma)
                                                      );
        }
    }

    float sumAccumulatorForFG = 0.0f;
    float marginalSumAccumulatorForFG = 0.0f;
    float weightAccumulatorForFG = 0.0f;

    const Pointer3D<unsigned char> fgMasksPtr(const_cast<unsigned char *>(fgMasks), imageDim.width, imageDim.height);

    for (int deltaT = 1; deltaT <= referenceFrames; ++deltaT) {
        /// referenceFrames < referenceBufferSize, hence:
        /// abs(inputIndex - deltaT) < config.referenceBufferSize ==> referenceIndex >= 0
        int const referenceIndex = ((inputIndex - deltaT) < 0)
                                   ? (inputIndex - deltaT + config.referenceBufferSize)
                                   : (inputIndex - deltaT);

        if (config.foregroundSpatialModelling) {
            weightAccumulatorForFG += static_cast<float> (numberOfFGSamples[referenceIndex]);
        }

        /// TODO : Por temas de coalescing y tal es más que probable que sea mejor, dado que ahora
        /// tenemos las imágenes como planos separados, que el bucle más exterior sea el de canales
        /// más que el espacial. Además sería bueno deshacerse del rollo de tener que usar
        /// arrayes para guardar la muestra original.
        /// Lamentablemente, el bucle interior es un productorio mientras que los exteriores
        /// son sumatorios. Además, mover los canales para fuera nos obligaría a leer la
        /// probabilidad una vez por canal en el caso de la actualización selectiva.


        for (int deltaY = -config.foregroundSpatialRadius;
                 deltaY <= config.foregroundSpatialRadius;
             ++deltaY) {
            for (int deltaX = -config.foregroundSpatialRadius;
                     deltaX <= config.foregroundSpatialRadius;
                 ++deltaX) {
                int const efX = x + deltaX;
                int const efY = y + deltaY;
                if (   (efX >= 0)             && (efY >= 0)
                    && (efX < imageDim.width) && (efY < imageDim.height)) {
                    // We only consider the sample if it was classified as FG
                    if (0 != fgMasksPtr(efX, efY, referenceIndex)) {
                        if (!config.foregroundSpatialModelling) {
                            weightAccumulatorForFG++;
                        }
                        float const spatialProduct = config.foregroundSpatialModelling
                                                     ? foregroundSpatialHorizontalGaussian[abs(deltaX)] * foregroundSpatialVerticalGaussian[abs(deltaY)]
                                                     : 1.0f; // Neutral element if we are not using spatial modelling
                        float partialProduct  = spatialProduct;
                        float const marginalProduct = spatialProduct;
                        float exponent = 0.0f;
                        #pragma unroll
                        for (int ch = 0; ch < CHANNELS; ++ch) {
                            //TODO: Si por alguna razón quisiéramos usar floats esto estaría mal
                            int const referenceImageSample = static_cast<int> (surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * efX, efY, referenceIndex * CHANNELS + ch));
                            exponent += static_cast<long>(referenceImageSample - currentImageSamples[ch])
                                        * static_cast<long>(referenceImageSample - currentImageSamples[ch])
                                        * invAppearanceSigmas[ch];
                        }
                        partialProduct *= __expf (exponent) * InvSqrt2Pi<CHANNELS>::VALUE;
                        //partialProduct *= negExp (-exponent);
                        sumAccumulatorForFG += partialProduct;
                        marginalSumAccumulatorForFG += marginalProduct;
                    }
                }
            }
        }
    }

    float const & alpha = config.foregroundMixingFactor;

    const float pdfForFG = (0.0f == weightAccumulatorForFG)
                           ? 1.0f
                           : (alpha + (1.0f - alpha) * sumAccumulatorForFG / (appearanceSigmasProduct * weightAccumulatorForFG))
                             / (config.spatialMarginalConditioning
                                ? (alpha + (1.0f - alpha) * marginalSumAccumulatorForFG /  weightAccumulatorForFG)
                                : 1.0f);



    *(displacePtr3D<float>(pdfFG,
                           imageDim.width,
                           imageDim.height,
                           x, y, 0)) = pdfForFG;
}

template <typename     ELEMENT_TYPE,
          unsigned int CHANNELS/*,
          bool         BACKGROUND_SPATIAL_MODELLING*/>
__global__ void computeAppearanceSigmas (ProcessingConfig                  config,
                                         ImageDimensions                   imageDim,
                                         int                               availablePastFrames,
                                         float const        * __restrict__ backgroundSpatialHorizontalGaussian,
                                         float const        * __restrict__ backgroundSpatialVerticalGaussian,
                                         long const         * __restrict__ imageSquaredDiffTemporalWeightedSum,
                                         long const         * __restrict__ probTemporalSum,
                                         float              * __restrict__ appearanceSigmas) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) {
        return; /// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    }
    Pointer3D<float> appearanceSigmasPtr(appearanceSigmas,   imageDim.width, imageDim.height);
    for (int ch = 0; ch < CHANNELS; ++ch) {
        appearanceSigmasPtr(x, y, ch) = ((availablePastFrames > 1)
                                && (availablePastFrames >= config.backgroundAppearanceSigmaEstimationDelay))
                               ? getAppearanceSigma<ELEMENT_TYPE,
                                                    CHANNELS/*,
                                                    BACKGROUND_SPATIAL_MODELLING*/> (config,
                                                                                   imageDim,
                                                                                   backgroundSpatialHorizontalGaussian,
                                                                                   backgroundSpatialVerticalGaussian,
                                                                                   imageSquaredDiffTemporalWeightedSum,
                                                                                   probTemporalSum,
                                                                                   ch)
                               : config.backgroundAppearanceSigma;
    }
    //((133 == x) && (133 == y)) ? printf("%f %f %f  --- ", appearanceSigmasPtr(x, y, 0), appearanceSigmasPtr(x, y, 1), appearanceSigmasPtr(x, y, 2)): 0;
}

template <unsigned int CHANNELS>
__global__ void weighMedianSigmas (ProcessingConfig                config,
                                   ImageDimensions                 imageDim,
                                   int                             availablePastFrames,
                                   float const      * __restrict__ backgroundSpatialHorizontalGaussian,
                                   float const      * __restrict__ backgroundSpatialVerticalGaussian,
                                   float const      * __restrict__ medianSigmas,
                                   float            * __restrict__ appearanceSigmas) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) {
        return; /// >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
    }
    const Pointer3D<float> medianSigmasPtr(const_cast<float *>(medianSigmas), imageDim.width, imageDim.height);
    Pointer3D<float> appearanceSigmasPtr(appearanceSigmas, imageDim.width, imageDim.height);

    if ((availablePastFrames > 1) && (availablePastFrames >= config.backgroundAppearanceSigmaEstimationDelay)) {
        for (int ch = 0; ch < CHANNELS; ++ch) {
            float sigmaWeightedSum = 0.0f;
            float weightsSum = 0.0f;
            for (int deltaY = -config.backgroundSpatialRadius;
                     deltaY <= config.backgroundSpatialRadius;
                 ++deltaY) {
                for (int deltaX = -config.backgroundSpatialRadius;
                         deltaX <= config.backgroundSpatialRadius;
                     ++deltaX) {
                    float const spatialWeight = config.backgroundSpatialModelling
                                                ? backgroundSpatialHorizontalGaussian[abs(deltaX)] * backgroundSpatialVerticalGaussian[abs(deltaY)]
                                                : 1.0f; // Neutral element if we are not using spatial modelling
                    weightsSum += spatialWeight;
                    int const readX = x + deltaX;
                    int const readY = y + deltaY;
                    if (   (readX >= 0)                && (readY >= 0)
                        && (readX < imageDim.width) && (readY < imageDim.height)) {
                        sigmaWeightedSum += spatialWeight * medianSigmasPtr(readX, readY, ch);
                    }
                }
            }
            appearanceSigmasPtr(x, y, ch) = __fdividef(sigmaWeightedSum, weightsSum);
        }
    } else {
        for (int ch = 0; ch < CHANNELS; ++ch) {
            appearanceSigmasPtr(x, y, ch) = config.backgroundAppearanceSigma;
        }
    }
}


template <typename ELEMENT_TYPE,
          unsigned int CHANNELS,
          bool BACKGROUND_SPATIAL_MODELLING,
          bool BACKGROUND_APPEARANCE_SIGMA_ESTIMATION,
          bool BACKGROUND_SELECTIVE_UPDATE>
__global__ void alternateProcessKernel (ProcessingConfig                  config,
                                        ImageDimensions                   imageDim,
                                        int                               availablePastFrames,
                                        int                               inputIndex,
                                        float const        * __restrict__ backgroundSpatialHorizontalGaussian,
                                        float const        * __restrict__ backgroundSpatialVerticalGaussian,
                                        unsigned char      * __restrict__ bgProbabilities,
                                        float const        * __restrict__ estimatedAppearanceSigmas,
                                        int const          * __restrict__ probSpatialSum,
                                        float              * __restrict__ pdfBG,
                                        unsigned char const      * __restrict__ thumbnail) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    if (config.roiAnalysis) {
        switch (thumbnail[blockIdx.x + blockIdx.y * gridDim.x]) {
        case Thumbnail::NOT_ANALYSED:
            // fallthrough
        case Thumbnail::POSITIVE:
            ;// fallthrough
        case Thumbnail::NEGATIVE:
            return;
        }
    }


    // Read current samples
    int currentImageSamples[CHANNELS];
    #pragma unroll
    for (int ch = 0; ch < CHANNELS; ++ch) {
    //TODO: Si por alguna razón quisiéramos usar floats esto estaría mal
        currentImageSamples[ch] = static_cast<int>(surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * x, y, inputIndex * CHANNELS + ch));
    }

    // Compute the number of available reference frames
    int const referenceFrames = min(availablePastFrames, config.backgroundReferenceFrames);

    float invAppearanceSigmas[CHANNELS];
    float appearanceSigmasProduct = 1.0f;
    {
        const Pointer3D<float> estimatedAppearanceSigmasPtr(const_cast<float *>(estimatedAppearanceSigmas), imageDim.width, imageDim.height);
        for (int ch = 0; ch < CHANNELS; ++ch) {
            const float currentAppearanceSigma = BACKGROUND_APPEARANCE_SIGMA_ESTIMATION ? estimatedAppearanceSigmasPtr(x, y, ch) : config.backgroundAppearanceSigma;
            appearanceSigmasProduct *= currentAppearanceSigma;

            invAppearanceSigmas[ch] = -0.5 / (static_cast<double>(ScalarInfo<ELEMENT_TYPE>::NORM_MAX)
                                                      * static_cast<double>(ScalarInfo<ELEMENT_TYPE>::NORM_MAX)
                                                      * static_cast<double>(currentAppearanceSigma)
                                                      * static_cast<double>(currentAppearanceSigma)
                                                      );
        }
    }

    float sumAccumulatorForBG = 0.0f;
    float marginalSumAccumulatorForBG = 0.0f;

    // Para actualización selectiva
    float weightAccumulatorForBG = BACKGROUND_SELECTIVE_UPDATE
                                   ? 2.22044604925031e-16//1E-9 // In case the rest of weights are zero
                                   : (referenceFrames * (BACKGROUND_SPATIAL_MODELLING
                                                         ? (imageDim.width * imageDim.height)
                                                         : 1));



    for (int deltaT = 1; deltaT <= referenceFrames; ++deltaT) {
        /// referenceFrames < referenceBufferSize, hence:
        /// abs(inputIndex - deltaT) < config.referenceBufferSize ==> referenceIndex >= 0
        int const referenceIndex = ((inputIndex - deltaT) < 0)
                                   ? (inputIndex - deltaT + config.referenceBufferSize)
                                   : (inputIndex - deltaT);

        float const temporalWeight = (BACKGROUND_SELECTIVE_UPDATE)
                                     ? ((deltaT <= config.backgroundSelectiveUpdateCutoff)
                                        ? 1.0f
                                        : computeG_T(config, deltaT))
                                     : 1.0f;

        if (!config.spatialMarginalConditioning) {
            if (BACKGROUND_SELECTIVE_UPDATE) {
                if (BACKGROUND_SPATIAL_MODELLING) {
                    // Las probabilidades son de momento unsigned char independientemente
                    // de ELEMENT_TYPE
                    weightAccumulatorForBG += temporalWeight * (static_cast<float> (probSpatialSum[referenceIndex]) / static_cast<float> (UCHAR_MAX))
                                              + ((1.0f - temporalWeight) * imageDim.width * imageDim.height);
                } else {
                    float const referenceProbBG = static_cast<float> (*(displacePtr3D<unsigned char>(bgProbabilities,
                                                                                                     imageDim.width,
                                                                                                     imageDim.height,
                                                                                                     x, y, referenceIndex)))
                                                  / static_cast<float> (UCHAR_MAX);
                    weightAccumulatorForBG += temporalWeight * referenceProbBG
                                              + 1.0f - temporalWeight;
                }
            }
        }


        /// TODO : Por temas de coalescing y tal es más que probable que sea mejor, dado que ahora
        /// tenemos las imágenes como planos separados, que el bucle más exterior sea el de canales
        /// más que el espacial. Además sería bueno deshacerse del rollo de tener que usar
        /// arrayes para guardar la muestra original.
        /// Lamentablemente, el bucle interior es un productorio mientras que los exteriores
        /// son sumatorios. Además, mover los canales para fuera nos obligaría a leer la
        /// probabilidad una vez por canal en el caso de la actualización selectiva.


        // Multiplying radiuses by BSP tells the compiler that the loop is not really a loop when it is disabled
        for (int deltaY = -config.backgroundSpatialRadius * static_cast<int> (BACKGROUND_SPATIAL_MODELLING);
                 deltaY <= config.backgroundSpatialRadius * static_cast<int> (BACKGROUND_SPATIAL_MODELLING);
             ++deltaY) {
            for (int deltaX = -config.backgroundSpatialRadius * static_cast<int> (BACKGROUND_SPATIAL_MODELLING);
                     deltaX <= config.backgroundSpatialRadius * static_cast<int> (BACKGROUND_SPATIAL_MODELLING);
                 ++deltaX) {
                int const efX = x + deltaX;
                int const efY = y + deltaY;
                if (   (efX >= 0)             && (efY >= 0)
                    && (efX < imageDim.width) && (efY < imageDim.height)) {
                    float const spatialProduct = (BACKGROUND_SPATIAL_MODELLING)
                                                 ? backgroundSpatialHorizontalGaussian[abs(deltaX)] * backgroundSpatialVerticalGaussian[abs(deltaY)]
                                                 : 1.0f; // Neutral element if we are not using spatial modelling
                    float partialProduct  = spatialProduct;
                    float marginalProduct = spatialProduct;
                    float exponent = 0.0f;
                    #pragma unroll
                    for (int ch = 0; ch < CHANNELS; ++ch) {
                        //TODO: Si por alguna razón quisiéramos usar floats esto estaría mal
                        int const referenceImageSample = static_cast<int> (surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * efX, efY, referenceIndex * CHANNELS + ch));
                        // TODO: Esto es para poder debuggear y comparar con lo del año pasado. Luego se quita.
//                        float const indexGaussian = __fdividef (abs (referenceImageSample - currentImageSamples[ch]) * (sndElements - 1),
//                                                                appearanceSigmas[ch] * sndCutoff * ScalarInfo<ELEMENT_TYPE>::NORM_MAX);
//                        partialProduct *= (indexGaussian > static_cast<float>(sndElements - 1)) ? 0 : tex1D (unifOrdinatesTex, 0.5f + indexGaussian);
//                        partialProduct *= standardNormal (__fdividef(referenceImageSample - currentImageSamples[ch], appearanceSigmas[ch] * static_cast<float>(ScalarInfo<ELEMENT_TYPE>::NORM_MAX)));
                        exponent += static_cast<long>(referenceImageSample - currentImageSamples[ch])
                                    * static_cast<long>(referenceImageSample - currentImageSamples[ch])
                                    * invAppearanceSigmas[ch];
                    }
                    partialProduct *= __expf (exponent) * InvSqrt2Pi<CHANNELS>::VALUE;
                    //partialProduct *= negExp (-exponent);
                    if (BACKGROUND_SELECTIVE_UPDATE) {
                        float const referenceProbBG = static_cast<float> (*(displacePtr3D<unsigned char>(bgProbabilities,
                                                                                                         imageDim.width,
                                                                                                         imageDim.height,
                                                                                                         efX, efY, referenceIndex)))
                                                      / static_cast<float> (UCHAR_MAX);
                        float const pixelWeight = temporalWeight * referenceProbBG + 1.0f - temporalWeight;
                        partialProduct *= pixelWeight;
                        marginalProduct *= pixelWeight;
                    }
                    sumAccumulatorForBG += partialProduct;
                    marginalSumAccumulatorForBG += marginalProduct;
                }
            }
        }
    }

    const float pdfForBG =   (0 == availablePastFrames)
                           ? FLT_MAX // To ensure that the first frame is fully classified as background
                           : (sumAccumulatorForBG
                              / (appearanceSigmasProduct * (config.spatialMarginalConditioning
                                                            ? marginalSumAccumulatorForBG
                                                            : weightAccumulatorForBG)));




    *(displacePtr3D<float>(pdfBG,
                           imageDim.width,
                           imageDim.height,
                           x, y, 0)) = pdfForBG;
}


template <typename ELEMENT_TYPE,
          unsigned int CHANNELS>
__global__ void updateHistogramsKernel (ProcessingConfig                config,
                                        ImageDimensions                 imageDim,
                                        int                             availablePastFrames,
                                        int                             inputIndex,
                                        unsigned char    * __restrict__ binsHistory,
                                        unsigned short   * __restrict__ diffsHistogram,
                                        float            * __restrict__ medianSigmas) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;

    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    int const newestReferenceDelta = 1;
    int const newestReferenceIndex = ( (inputIndex - newestReferenceDelta) < 0)
                                     ? (inputIndex - newestReferenceDelta + config.referenceBufferSize)
                                     : (inputIndex - newestReferenceDelta);
    /// Cosas que no dependen del número de canales:
    /// - Comprobar si tenemos suficientes imágenes de referencia como
    ///   para tener que considerar las cosas caducadas.
    bool const fullReference = (availablePastFrames >= config.backgroundReferenceFrames);
    int const oldestReferenceDelta = config.backgroundReferenceFrames;
    int const oldestReferenceIndex = ( (inputIndex - oldestReferenceDelta) < 0)
                                     ? (inputIndex - oldestReferenceDelta + config.referenceBufferSize)
                                     : (inputIndex - oldestReferenceDelta);

    Pointer4D<unsigned char> binsHistoryPtr(binsHistory, imageDim.width, imageDim.height, CHANNELS);
    Pointer4D<unsigned short> diffsHistogramPtr(diffsHistogram, imageDim.width, imageDim.height, NUMBINS);
    Pointer3D<float> medianSigmasPtr(medianSigmas, imageDim.width, imageDim.height);
    for (int ch = 0; ch < CHANNELS; ++ch) {
        /// - Leer muestra de referencia más reciente y muestra actual si no las
        ///   tenemos ya.
        long const currentImageSample   = static_cast<long> (surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * x, y, inputIndex * CHANNELS + ch));
        long const referenceImageSample = (availablePastFrames > 0)
                                          ? static_cast<long> (surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * x, y, newestReferenceIndex * CHANNELS + ch))
                                          : currentImageSample; // If we have no reference, we want the difference to be zero

        /// - Calcular la nueva diferencia y añadirla al histograma y a la historia
        float const newImageAbsDiff = __fdividef(labs(currentImageSample - referenceImageSample), ScalarInfo<ELEMENT_TYPE>::NORM_MAX);
        const unsigned char newBin = float2bin(newImageAbsDiff);
        ++diffsHistogramPtr(x, y, newBin, ch);
        binsHistoryPtr(x, y, ch, inputIndex) = newBin;
        /// - Retirar del histograma las muestras caducadas si procede
        if (fullReference) {
            --diffsHistogramPtr(x, y, binsHistoryPtr(x, y, ch, oldestReferenceIndex), ch);
        }
//        unsigned int histogramSum = 0;
//        for (unsigned int b = 0; b < NUMBINS; ++b) {
//            histogramSum += diffsHistogramPtr(x, y, b, ch);
//        }
//        if (histogramSum != (fullReference ? (config.backgroundReferenceFrames) : (availablePastFrames + 1))) { // La primera vez incorporamos una diferencia nula espuria
//            printf("||| x %d y %d newbin %d diff %f sum %d past %d ", x, y, newBin, newImageAbsDiff, histogramSum, availablePastFrames);
//            asm("trap;");
//        }
        const unsigned int histogramSum = fullReference ? (config.backgroundReferenceFrames) : (availablePastFrames + 1);
        unsigned int runningSum = 0;
        for (unsigned int b = 0; b < NUMBINS; ++b) {
            runningSum += diffsHistogramPtr(x, y, b, ch);
            if (runningSum >= (histogramSum / 2)) {
                medianSigmasPtr(x, y, ch) = bin2float(b) * 1.039862914f; //     = / (0.68 * sqrt(2)), (5.25 en tesis ccr)
                break;
            }
        }
    }
}




template <typename ELEMENT_TYPE,
          unsigned int CHANNELS>
__global__ void updateAppearanceSigmaEstimationKernel (ProcessingConfig  config,
                                                       ImageDimensions   imageDim,
                                                       int               availablePastFrames,
                                                       int               inputIndex,
                                                       ELEMENT_TYPE     *__restrict__ imageAbsDiffs,
                                                       unsigned char    *__restrict__ bgProbabilities,
                                                       long             *__restrict__ imageSquaredDiffTemporalWeightedSum,
                                                       long             *__restrict__ probTemporalSum
                                                       )
{
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    int const newestReferenceDelta = 1;
    int const newestReferenceIndex = ( (inputIndex - newestReferenceDelta) < 0)
                                     ? (inputIndex - newestReferenceDelta + config.referenceBufferSize)
                                     : (inputIndex - newestReferenceDelta);
    /// Cosas que no dependen del número de canales:
    /// - Comprobar si tenemos suficientes imágenes de referencia como
    ///   para tener que considerar las cosas caducadas.
    bool const fullReference = (availablePastFrames >= config.backgroundReferenceFrames);
    int const oldestReferenceDelta = config.backgroundReferenceFrames;
    int const oldestReferenceIndex = ( (inputIndex - oldestReferenceDelta) < 0)
                                     ? (inputIndex - oldestReferenceDelta + config.referenceBufferSize)
                                     : (inputIndex - oldestReferenceDelta);
    /// - Leer probabilidad caducada si procede, 0 en caso contrario.
    long const expiredProbBG = fullReference
                               ? static_cast<long> (*(displacePtr3D<unsigned char>(bgProbabilities,
                                                                                   imageDim.width,
                                                                                   imageDim.height,
                                                                                   x, y, oldestReferenceIndex)))
                                                    : 0;
    /// - Leer (si no tenemos ya la probabilidad nueva). Comprobar que en
    ///   esta versión se está escribiendo correctamente en su sitio. En el
    ///   futuro esto se deberá hacer en la parte de actualización, imagino.
    long const newProbBG = static_cast<long> (*(displacePtr3D<unsigned char>(bgProbabilities,
                                                                             imageDim.width,
                                                                             imageDim.height,
                                                                             x, y, inputIndex)));
    /// - Leer la probabilidad total antigua
    long const oldProbTemporalSum = *(displacePtr3D<long>(probTemporalSum,
                                                          imageDim.width,
                                                          imageDim.height,
                                                          x, y,
                                                          0));
    /// - Calcular probabilidad total nueva restando la probabilidad
    ///   caducada y añadiendo la nueva y guardarla.
    long const newProbTemporalSum = oldProbTemporalSum + newProbBG - expiredProbBG;
    *(displacePtr3D<long>(probTemporalSum,
                          imageDim.width,
                          imageDim.height,
                          x, y,
                          0)) = newProbTemporalSum;

    ///
    /// Cosas que se hacen por cada canal:
    for (int ch = 0; ch < CHANNELS; ++ch) {
    /// - Leer muestra de referencia más reciente y muestra actual si no las
    ///   tenemos ya.
        long const currentImageSample   = static_cast<long> (surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * x, y, inputIndex * CHANNELS + ch));
        long const referenceImageSample = (availablePastFrames > 0)
                                          ? static_cast<long> (surf2DLayeredread<ELEMENT_TYPE> (inputImageSurface, sizeof (ELEMENT_TYPE) * x, y, newestReferenceIndex * CHANNELS + ch))
                                          : currentImageSample; // If we have no reference, we want the difference to be zero

    /// - Calcular la nueva diferencia y guardarla
        long const newImageAbsDiff = labs(currentImageSample - referenceImageSample);
        *(displacePtr3D<ELEMENT_TYPE>(imageAbsDiffs, imageDim.width, imageDim.height, x, y, ch + CHANNELS * inputIndex)) = static_cast<ELEMENT_TYPE> (newImageAbsDiff);
    /// - Leer la diferencia caducada si procede, 0 en caso contrario.
        long const expiredImageAbsDiff = fullReference
                                         ? static_cast<long> (*(displacePtr3D<ELEMENT_TYPE>(imageAbsDiffs, imageDim.width, imageDim.height, x, y, ch + CHANNELS * oldestReferenceIndex)))
                                         : 0;
    /// - Leer suma ponderada antigua
        long const oldImageSquaredDiffTemporalWeightedSum = *(displacePtr3D<long>(imageSquaredDiffTemporalWeightedSum,
                                                                                  imageDim.width,
                                                                                  imageDim.height,
                                                                                  x, y,
                                                                                  ch));

    /// - Calcular suma ponderada nueva retirando el producto de la
    ///   probabilidad caducada por el cuadrado de la diferencia caducada y
    ///   añadiendo el producto de la probabilidad nueva
        long const newImageSquaredDiffTemporalWeightedSum = oldImageSquaredDiffTemporalWeightedSum
                                                            + newImageAbsDiff * newImageAbsDiff * newProbBG
                                                            - expiredImageAbsDiff * expiredImageAbsDiff * expiredProbBG;
        *(displacePtr3D<long>(imageSquaredDiffTemporalWeightedSum,
                              imageDim.width,
                              imageDim.height,
                              x, y,
                              ch)) = newImageSquaredDiffTemporalWeightedSum;
    }



}





__global__ void classifierKernel (ProcessingConfig                config,
                                  ImageDimensions                 imageDim,
                                  int                             inputIndex,
                                  int                             outputIndex,
                                  unsigned char    * __restrict__ bgProbabilities,
                                  float const      * __restrict__ pdfFG,
                                  float const      * __restrict__ pdfBG,
                                  float const      * __restrict__ fgPriorProbability,
                                  unsigned char    * __restrict__ fgMask,
                                  unsigned char    * __restrict__ thumbnail) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    Pointer3D<unsigned char> bgProbabilities3DPtr(bgProbabilities, imageDim.width, imageDim.height);
    Pointer2D<float>         fgPrior2DPtr(const_cast<float *>(fgPriorProbability), imageDim.width);


    if (config.roiAnalysis) {
        switch (thumbnail[blockIdx.x + blockIdx.y * gridDim.x]) {
        case Thumbnail::NOT_ANALYSED:
            surf2DLayeredwrite<float> (0.0f, nextStageSurface, sizeof (float) * x, y, outputIndex);
//            *(displacePtr3D<unsigned char>(bgProbabilities,
//                                           imageDim.width,
//                                           imageDim.height,
//                                           x, y, inputIndex)) = UCHAR_MAX;
            bgProbabilities3DPtr(x, y, inputIndex) = UCHAR_MAX;
            *(displacePtr3D<unsigned char>(fgMask,
                                           imageDim.width,
                                           imageDim.height,
                                           x, y, 0)) = static_cast<unsigned char>(0);
            // fallthrough
        case Thumbnail::POSITIVE:
            ;// fallthrough
        case Thumbnail::NEGATIVE:
            return;
        }
    }


    float const bg = *(displacePtr3D<float>(pdfBG,
                                            imageDim.width,
                                            imageDim.height,
                                            x, y, 0)) * (config.particleFilterEnabled ? 1.0f - fgPrior2DPtr(x,y) : 1.0f);
/// Análisis bayesiano habitual
    float const fg = *(displacePtr3D<float>(pdfFG,
                                            imageDim.width,
                                            imageDim.height,
                                            x, y, 0)) * (config.particleFilterEnabled ? fgPrior2DPtr(x,y) : 1.0f);
    float const probFG = ((x < config.backgroundSpatialRadius)
                          || (y < config.backgroundSpatialRadius)
                          || (x >= imageDim.width - config.backgroundSpatialRadius)
                          || (y >= imageDim.height - config.backgroundSpatialRadius))
                         ? 0.0f : (fg / (fg + bg));

/// Comparación simple con un umbral
//    float const probFG = static_cast<float>(bg < 1e0);




    *(displacePtr3D<unsigned char>(fgMask,
                                   imageDim.width,
                                   imageDim.height,
                                   x, y, 0)) = static_cast<unsigned char>(probFG > 0.5f);
    surf2DLayeredwrite<float> (probFG, nextStageSurface, sizeof (float) * x, y, outputIndex);
//    surf2DLayeredwrite<float> (fgPrior2DPtr(x,y), nextStageSurface, sizeof (float) * x, y, outputIndex);


    /// We write the prob(BG) here because even if we are not using appearance
    /// sigma estimation or dynamic update, we will need it for the foreground
    /// model.
    unsigned char const probBG = static_cast<unsigned char> (rintf ((1.0f - probFG) * UCHAR_MAX));
//    *(displacePtr3D<unsigned char>(bgProbabilities,
//                                   imageDim.width,
//                                   imageDim.height,
//                                   x, y, inputIndex)) = probBG;
    bgProbabilities3DPtr(x, y, inputIndex) = probBG;

    if (config.roiAnalysis) {
        /// Now we will reduce the results of the classifier to decide whether
        /// we have detected enough FG to consider expanding analysis to its
        /// neighbours.
        extern __shared__ unsigned short detected[];
        unsigned int linearId = threadIdx.x + threadIdx.y * blockDim.x;
        detected[linearId] = static_cast<unsigned short> (probBG < (UCHAR_MAX / 2));
        __syncthreads ();
        // We will now be progressively halving the number of threads
        // that will take part in the computation until there is only one
        for (int activeThreads = blockDim.x * blockDim.y / 2;
             activeThreads > 0;
             activeThreads /= 2) {
            if (linearId < activeThreads) {
                // Each active thread adds a position written by a thread
                // that was active in the previous iteration but not anymore
                // to its own position.
                detected[linearId] += detected[linearId + activeThreads];
            }
            if (activeThreads > WARP_SIZE) { // If there is still more than one warp active
                __syncthreads (); // we need to ensure proper synchronization.
            } // If there is only one active warp, it is synchronized by definition.
        }
        if (0 == linearId) {
            // We set to one the pixel representing this block if more than 1/10 samples are positive
            thumbnail[blockIdx.x + blockIdx.y * gridDim.x] = ((10 * detected[0]) > (blockDim.x * blockDim.y))
                                                             ? Thumbnail::POSITIVE
                                                             : Thumbnail::NEGATIVE;
        }
    }
}


__global__ void dilateSquare (ImageDimensions                    imageDim,
                              const unsigned char * __restrict__ input,
                              unsigned char       * __restrict__ output,
                              unsigned int                       radius) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    int const intRadius = static_cast<int>(radius);
    int const loX = max(x - intRadius, 0);
    int const hiX = min(x + intRadius, imageDim.width  - 1);
    int const loY = max(y - intRadius, 0);
    int const hiY = min(y + intRadius, imageDim.height - 1);

    unsigned char outputPixel = 0;

    for (int refY = loY; refY <= hiY; ++refY) {
        for (int refX = loX; refX <= hiX; ++refX) {
            if (0 < *(displacePtr3D<unsigned char>(input,
                                                   imageDim.width,
                                                   imageDim.height,
                                                   refX, refY, 0))) {
                outputPixel = 1;
            }
        }
    }

    *(displacePtr3D<unsigned char>(output,
                                   imageDim.width,
                                   imageDim.height,
                                   x, y, 0)) = outputPixel;
}

__global__ void erodeSquare (ImageDimensions                    imageDim,
                             const unsigned char * __restrict__ input,
                             unsigned char       * __restrict__ output,
                             unsigned int                       radius) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    int const intRadius = static_cast<int>(radius);
    int const loX = max(x - intRadius, 0);
    int const hiX = min(x + intRadius, imageDim.width  - 1);
    int const loY = max(y - intRadius, 0);
    int const hiY = min(y + intRadius, imageDim.height - 1);

    unsigned char outputPixel = 1;

    for (int refY = loY; refY <= hiY; ++refY) {
        for (int refX = loX; refX <= hiX; ++refX) {
            if (0 == *(displacePtr3D<unsigned char>(input,
                                                    imageDim.width,
                                                    imageDim.height,
                                                    refX, refY, 0))) {
                outputPixel = 0;
            }
        }
    }

    *(displacePtr3D<unsigned char>(output,
                                   imageDim.width,
                                   imageDim.height,
                                   x, y, 0)) = outputPixel;
}

__global__ void dilateCircle (ImageDimensions                    imageDim,
                              const unsigned char * __restrict__ input,
                              unsigned char       * __restrict__ output,
                              unsigned int                       radius) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    int const intRadius = static_cast<int>(radius);
    int const loX = max(x - intRadius, 0);
    int const hiX = min(x + intRadius, imageDim.width  - 1);
    int const loY = max(y - intRadius, 0);
    int const hiY = min(y + intRadius, imageDim.height - 1);

    unsigned char outputPixel = 0;

    for (int refY = loY; refY <= hiY; ++refY) {
        for (int refX = loX; refX <= hiX; ++refX) {
            if (radius * radius >= ((refX - x) * (refX - x) + (refY - y) * (refY - y))) {
                continue;
            }
            if (0 < *(displacePtr3D<unsigned char>(input,
                                                   imageDim.width,
                                                   imageDim.height,
                                                   refX, refY, 0))) {
                outputPixel = 1;
            }
        }
    }

    *(displacePtr3D<unsigned char>(output,
                                   imageDim.width,
                                   imageDim.height,
                                   x, y, 0)) = outputPixel;
}

__global__ void erodeCircle (ImageDimensions                    imageDim,
                             const unsigned char * __restrict__ input,
                             unsigned char       * __restrict__ output,
                             unsigned int                       radius) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    int const intRadius = static_cast<int>(radius);
    int const loX = max(x - intRadius, 0);
    int const hiX = min(x + intRadius, imageDim.width  - 1);
    int const loY = max(y - intRadius, 0);
    int const hiY = min(y + intRadius, imageDim.height - 1);

    unsigned char outputPixel = 1;

    for (int refY = loY; refY <= hiY; ++refY) {
        for (int refX = loX; refX <= hiX; ++refX) {
            if (radius * radius >= ((refX - x) * (refX - x) + (refY - y) * (refY - y))) {
                continue;
            }
            if (0 == *(displacePtr3D<unsigned char>(input,
                                                    imageDim.width,
                                                    imageDim.height,
                                                    refX, refY, 0))) {
                outputPixel = 0;
            }
        }
    }

    *(displacePtr3D<unsigned char>(output,
                                   imageDim.width,
                                   imageDim.height,
                                   x, y, 0)) = outputPixel;
}


__global__ void writeBinaryToOutput (ImageDimensions                    imageDim,
                                     int                                outputIndex,
                                     const unsigned char * __restrict__ fgMask) {
    int const x = blockIdx.x * blockDim.x + threadIdx.x;
    int const y = blockIdx.y * blockDim.y + threadIdx.y;
    if ((0 > x) || (0 > y) || (x >= imageDim.width) || (y >= imageDim.height)) { return; }

    surf2DLayeredwrite<float> (*(displacePtr3D<unsigned char>(fgMask,
                                                              imageDim.width,
                                                              imageDim.height,
                                                              x, y, 0)) > 0 ? 1.0f : 0.0f,
                               nextStageSurface,
                               sizeof (float) * x, y, outputIndex);
}


template <typename ELEMENT_TYPE>
struct BackgroundKernelDispatcher {
typedef void (*KernelPtr)(ProcessingConfig,
                                    ImageDimensions,
                                    int,
                                    int,
                                    float const *,
                                    float const *,
                                    unsigned char   *,
                                    float const *,
                                    int const        *,
                                    float *,
                                    unsigned char const *);

    template <unsigned int mode>
    void processingKernelModeInitializer () {
        this->ptrs[DecimalToBinaryBit<0, mode>::Value]
                  [DecimalToBinaryBit<1, mode>::Value]
                  [DecimalToBinaryBit<2, mode>::Value] = alternateProcessKernel <ELEMENT_TYPE,
                                                                                 3,
                                                                                 DecimalToBinaryBit<0, mode>::Value,
                                                                                 DecimalToBinaryBit<1, mode>::Value,
                                                                                 DecimalToBinaryBit<2, mode>::Value>;
    }

    template <unsigned int mode>
    void processingKernelInitializeAllModes () {
        processingKernelModeInitializer<mode> ();
        processingKernelInitializeAllModes<mode - 1> ();
    }

    BackgroundKernelDispatcher () {
        processingKernelInitializeAllModes<7> ();
    }


    KernelPtr ptrs[2][2][2];
};

template <>
template <>
void BackgroundKernelDispatcher<unsigned short>::processingKernelInitializeAllModes<0> () {
    processingKernelModeInitializer<0> ();
}

template <>
template <>
void BackgroundKernelDispatcher<unsigned char>::processingKernelInitializeAllModes<0> () {
    processingKernelModeInitializer<0> ();
}







template<typename T>
void launchProcess (const ProcessingConfig    & config,
                    const ImageDimensions     & imageDim,
                    const ProcessingDeviceAux & deviceAux,
                    ProcessingDeviceData<T>   & deviceData,
                    ProcessingHostData        & hostData) {

    static const BackgroundKernelDispatcher<T> dispatcher;
    static const typename BackgroundKernelDispatcher<T>::KernelPtr actualBGKernel = dispatcher.ptrs[config.backgroundSpatialModelling         ]
                                                                                                   [config.backgroundAppearanceSigmaEstimation]
                                                                                                   [config.backgroundSelectiveUpdate          ];
    static ForegroundReferenceData<T, 3> fgData(imageDim, 10);

    static T * d_MultiSampleImage = NULL;
    static unsigned int * d_SampleCount = NULL;

    hostData.thumbnail.addRandom(1, 4, 4);

    hostData.particleFilter.getPriorForegroundProbability(deviceData.fgPriorProbability);

    Pointer3D<unsigned char> d_fgMaskPtr(deviceData.fgMasks, imageDim.width, imageDim.height);

    do {
        //assert(hostData.thumbnail.writeImage());
        hostData.thumbnail.toDevice (hostData.stream);
        if (config.backgroundAppearanceSigmaEstimation) {
//            computeAppearanceSigmas<T, 3><<<hostData.gridSize,
//                                            hostData.blockSize,
//                                            0,
//                                            hostData.stream>>>(config,
//                                                               imageDim,
//                                                               hostData.availableReference,
//                                                               deviceAux.backgroundSpatialHorizontalGaussian,
//                                                               deviceAux.backgroundSpatialVerticalGaussian,
//                                                               deviceData.imageSquaredDiffTemporalWeightedSum,
//                                                               deviceData.probTemporalSum,
//                                                               deviceData.backgroundAppearanceSigmas);
            weighMedianSigmas<3><<<hostData.gridSize,
                                   hostData.blockSize,
                                   0,
                                   hostData.stream>>>(config,
                                                      imageDim,
                                                      hostData.availableReference,
                                                      deviceAux.backgroundSpatialHorizontalGaussian,
                                                      deviceAux.backgroundSpatialVerticalGaussian,
                                                      deviceData.medianSigmas,
                                                      deviceData.backgroundAppearanceSigmas);
        }
        assert (NULL != actualBGKernel);
        actualBGKernel <<<hostData.gridSize,
                          hostData.blockSize,
                          0,
                          hostData.stream>>> (config,
                                     imageDim,
                                     hostData.availableReference,
                                     hostData.readIndex,
                                     deviceAux.backgroundSpatialHorizontalGaussian,
                                     deviceAux.backgroundSpatialVerticalGaussian,
                                     deviceData.bgProbabilities,
                                     deviceData.backgroundAppearanceSigmas,
                                     deviceData.probSpatialSum,
                                     deviceData.pdfBG,
                                     hostData.thumbnail.getDeviceData ());

        if (config.foregroundModelEnabled && (hostData.availableReference >= config.foregroundStartDelay)) {
            if (config.particleFilterEnabled) {
                multiSampleForegroundModelKernel<T, 3> <<<hostData.gridSize,
                                            hostData.blockSize,
                                            0,
                                            hostData.stream>>>(config,
                                                               imageDim,
                                                               hostData.availableReference,
                                                               hostData.readIndex,
                                                               deviceAux.foregroundSpatialHorizontalGaussian,
                                                               deviceAux.foregroundSpatialVerticalGaussian,
                                                               deviceData.numberOfFGSamples,
                                                               d_MultiSampleImage,
                                                               d_SampleCount,
                                                               deviceData.backgroundAppearanceSigmas,
                                                               deviceData.pdfFG,
                                                               hostData.thumbnail.getDeviceData ());

            } else {
                foregroundModelKernel<T, 3> <<<hostData.gridSize,
                                            hostData.blockSize,
                                            0,
                                            hostData.stream>>>(config,
                                                               imageDim,
                                                               hostData.availableReference,
                                                               hostData.readIndex,
                                                               deviceAux.foregroundSpatialHorizontalGaussian,
                                                               deviceAux.foregroundSpatialVerticalGaussian,
                                                               deviceData.numberOfFGSamples,
                                                               deviceData.fgMasks,
                                                               deviceData.backgroundAppearanceSigmas,
                                                               deviceData.pdfFG,
                                                               hostData.thumbnail.getDeviceData ());
            }
        } else {
            uniformForegroundModelKernel<<<hostData.gridSize,
                                           hostData.blockSize,
                                           0,
                                           hostData.stream>>> (config, imageDim, deviceData.pdfFG, hostData.thumbnail.getDeviceData ());
        }



        classifierKernel <<<hostData.gridSize,
                            hostData.blockSize,
                            static_cast<int> (config.roiAnalysis) * hostData.blockSize.x * hostData.blockSize.y * sizeof (unsigned short), // TODO: ¿Qué es este ushort?
                            hostData.stream>>> (config,
                                       imageDim,
                                       hostData.readIndex,
                                       hostData.writeIndex,
                                       deviceData.bgProbabilities,
                                       deviceData.pdfFG,
                                       deviceData.pdfBG,
                                       deviceData.fgPriorProbability,
                                       deviceData.morphologicalFG,
                                       hostData.thumbnail.getDeviceData ());
        hostData.thumbnail.fromDevice (hostData.stream);
        CUDA_SAFE_CALL (cudaStreamSynchronize (hostData.stream));
        CUDA_SAFE_CALL (cudaGetLastError());
    } while (hostData.thumbnail.nextIteration (2));
    //assert(hostData.thumbnail.writeImage());

    Pointer3D<unsigned char> d_morphologicalFGPtr(deviceData.morphologicalFG, imageDim.width, imageDim.height);
    if (config.morphologicalFilter) {
        dilateCircle<<<hostData.gridSize, hostData.blockSize, 0, hostData.stream>>> (imageDim, &d_morphologicalFGPtr(0, 0, 0), &d_morphologicalFGPtr(0, 0, 1), 3);
        erodeCircle<<<hostData.gridSize, hostData.blockSize, 0, hostData.stream>>> (imageDim, &d_morphologicalFGPtr(0, 0, 1), &d_morphologicalFGPtr(0, 0, 4), 3);

        /// OJO: Hemos dejado sólo el cierre para juntar zonas adyacentes, sin nada de lo que pueda perjudicar esto, como es la erosión final o la apertura
        /// Para escribir en el sitio correcto vamos directos a (0,0,5) tras la erosión del cierre. Ojo al volverlo a poner como estaba.
        /// 2 . Ahora hemos reactivado la erosión final a ver si le mola más.



//        erode3x3<<<hostData.gridSize, hostData.blockSize, 0, hostData.stream>>> (imageDim, &d_morphologicalFGPtr(0, 0, 1), &d_morphologicalFGPtr(0, 0, 2));
//        erode3x3<<<hostData.gridSize, hostData.blockSize, 0, hostData.stream>>> (imageDim, &d_morphologicalFGPtr(0, 0, 2), &d_morphologicalFGPtr(0, 0, 3));
//        dilate3x3<<<hostData.gridSize, hostData.blockSize, 0, hostData.stream>>> (imageDim, &d_morphologicalFGPtr(0, 0, 3), &d_morphologicalFGPtr(0, 0, 4));
        erodeCircle<<<hostData.gridSize, hostData.blockSize, 0, hostData.stream>>> (imageDim, &d_morphologicalFGPtr(0, 0, 4), &d_morphologicalFGPtr(0, 0, 5), 1);
    }
    cudaMemcpyAsync(&d_fgMaskPtr(0, 0, hostData.readIndex),
                    &d_morphologicalFGPtr(0, 0, config.morphologicalFilter ? 5 : 0),
                    imageDim.width * imageDim.height * sizeof(unsigned char),
                    cudaMemcpyDeviceToDevice, hostData.stream);


    if (config.backgroundAppearanceSigmaEstimation) {
//        updateAppearanceSigmaEstimationKernel<T, 3> <<<hostData.gridSize, hostData.blockSize, 0, hostData.stream>>> (config,
//                                                                                       imageDim,
//                                                                                       hostData.availableReference,
//                                                                                       hostData.readIndex,
//                                                                                       deviceData.imageAbsDiffs,
//                                                                                       deviceData.bgProbabilities,
//                                                                                       deviceData.imageSquaredDiffTemporalWeightedSum,
//                                                                                       deviceData.probTemporalSum);
//        CUDA_SAFE_CALL (cudaStreamSynchronize (hostData.stream));
//        CUDA_SAFE_CALL (cudaGetLastError());
        updateHistogramsKernel<T, 3> <<<hostData.gridSize, hostData.blockSize, 0, hostData.stream>>> (config,
                                                                                                      imageDim,
                                                                                                      hostData.availableReference,
                                                                                                      hostData.readIndex,
                                                                                                      deviceData.binsHistory,
                                                                                                      deviceData.diffsHistograms,
                                                                                                      deviceData.medianSigmas);
//                                                                                                      deviceData.backgroundAppearanceSigmas);
        CUDA_SAFE_CALL (cudaStreamSynchronize (hostData.stream));
        CUDA_SAFE_CALL (cudaGetLastError());
    }

    if (config.backgroundSelectiveUpdate && config.backgroundSpatialModelling) {
        thrust::device_ptr<unsigned char> bgProbabilitiesPtr (deviceData.bgProbabilities);
        thrust::device_ptr<int> probSpatialSumPtr (deviceData.probSpatialSum);
        probSpatialSumPtr[hostData.readIndex] =  thrust::reduce (bgProbabilitiesPtr + hostData.readIndex * imageDim.width * imageDim.height,
                                                         bgProbabilitiesPtr + (1 + hostData.readIndex) * imageDim.width * imageDim.height,
                                                         static_cast<int> (0),
                                                         thrust::plus<int>()); // This forces the promotion to int before addition to avoid overflow
    }
    if (config.foregroundModelEnabled && config.foregroundSpatialModelling) {
        thrust::device_ptr<unsigned char> bgProbabilitiesPtr (deviceData.bgProbabilities);
        thrust::device_ptr<int> numberOfFGSamplesPtr (deviceData.numberOfFGSamples);
        numberOfFGSamplesPtr[hostData.readIndex] = thrust::transform_reduce (bgProbabilitiesPtr + hostData.readIndex * imageDim.width * imageDim.height,
                                                                     bgProbabilitiesPtr + (1 + hostData.readIndex) * imageDim.width * imageDim.height,
                                                                     LessThan128<int> (),
                                                                     static_cast<int> (0),
                                                                     thrust::plus<int> ());
    }

    if (config.particleFilterEnabled) {
        hostData.particleFilter.processFrame(&d_fgMaskPtr(0, 0, hostData.readIndex), displacePtr3D(deviceData.bgProbabilities, imageDim.width, imageDim.height, 0, 0, hostData.readIndex));

        fgData.insertNewFrame(hostData.particleFilter.getCurrentLabelledPixels(), hostData.readIndex);
        fgData.updatePixelLocations(hostData.particleFilter.getStateVectors());

        fgData.uploadToDeviceAsImage(&d_MultiSampleImage, &d_SampleCount, hostData.stream, false); // Last parameter true to show data count /run/shm/datacount
        fgData.initializeThumbnail(hostData.thumbnail.getHostData(), hostData.gridSize, hostData.blockSize);
    } else {
        hostData.thumbnail.initializeNextFrame();
    }
    hostData.thumbnail.advanceFrame();
}

// Explicit instantiation
template
void launchProcess<unsigned char> (ProcessingConfig const &config,
                    ImageDimensions  const &imageDim,
                    ProcessingDeviceAux const & deviceAux,
                    ProcessingDeviceData<unsigned char>  &deviceData,
                    ProcessingHostData     & hostData);

template
void launchProcess<unsigned short> (ProcessingConfig const &config,
                    ImageDimensions  const &imageDim,
                    ProcessingDeviceAux const & deviceAux,
                    ProcessingDeviceData<unsigned short>  &deviceData,
ProcessingHostData     & hostData);

template <typename T, unsigned int CHANNELS>
__global__ void updatePixelColoursKernel (ColouredPixel<T, CHANNELS> * d_pixels, unsigned int d_pixelsLength, unsigned int referenceImageIndex) {
    const unsigned int x = blockIdx.x * blockDim.x + threadIdx.x;

    if (x >= d_pixelsLength) {
        return;
    }
    ColouredPixel<T, CHANNELS> & px = d_pixels[x];
    for (unsigned int ch = 0; ch < CHANNELS; ++ch) {
        px.colour[ch] = surf2DLayeredread<T> (inputImageSurface, sizeof (T) * px.col, px.row, referenceImageIndex * CHANNELS + ch);
    }
}

// We assume d_pixels is big enough to hold h_pixels
template <typename T, unsigned int CHANNELS>
void updatePixelColours (ColouredPixel<T, CHANNELS> * d_pixels, std::vector<ColouredPixel<T, CHANNELS> > & h_pixels, unsigned int referenceImageIndex, cudaStream_t stream) {
    if (0 < h_pixels.size()) {
        CUDA_SAFE_CALL(cudaMemcpyAsync(d_pixels, &(h_pixels.at(0)), sizeof(ColouredPixel<T, CHANNELS>) * h_pixels.size(), cudaMemcpyHostToDevice, stream));
        updatePixelColoursKernel<T,CHANNELS><<<ceilUintDivision(h_pixels.size(), 256), 256, 0, stream>>>(d_pixels, h_pixels.size(), referenceImageIndex);
        CUDA_SAFE_CALL(cudaMemcpyAsync(&(h_pixels.at(0)), d_pixels, sizeof(ColouredPixel<T, CHANNELS>) * h_pixels.size(), cudaMemcpyDeviceToHost, stream));
        CUDA_SAFE_CALL(cudaStreamSynchronize(stream));
    }
}

template void updatePixelColours<unsigned char, 3> (ColouredPixel<unsigned char, 3> * d_pixels,
                                                    std::vector<ColouredPixel<unsigned char, 3> > & h_pixels,
                                                    unsigned int referenceImageIndex, cudaStream_t stream);

template void updatePixelColours<unsigned short, 3> (ColouredPixel<unsigned short, 3> * d_pixels,
                                                    std::vector<ColouredPixel<unsigned short, 3> > & h_pixels,
                                                    unsigned int referenceImageIndex, cudaStream_t stream);




